#! /usr/bin/env python
# Copyright (c) 2021 Predibase, Inc.

import logging
import os
from typing import Any, Dict, Optional

import pandas as pd

from predibase.pql.api import Session
from predibase.pql.utils import remove_trailing_slash
from predibase.util import DEFAULT_API_ENDPOINT, DEFAULT_SERVING_ENDPOINT, log_info
from predibase.util.settings import load_settings

_session = Session()
logger = logging.getLogger(__name__)


def start_session(gateway: str = None, token: str = None, serving_endpoint: str = None) -> Session:
    # Always starts a new session
    session = Session()
    if session.url is None:
        connect_session(session, url=gateway, serving_endpoint=serving_endpoint, token=token)

    # Return message upon successful client connection
    try:
        user = session.get_current_user()

        if user is not None:
            log_info(f"Connected to Predibase as {user}")

    except Exception as e:
        raise ValueError(f"Unable to connect to Predibase: {e}")

    return session


def connect_session(
    session: Session,
    url: Optional[str] = None,
    serving_endpoint: Optional[str] = None,
    token: Optional[str] = None,
    verbose: Optional[bool] = None,
):
    settings = load_settings()
    token = token or settings.get("token")
    url = url or settings.get("endpoint")

    if url is None:
        url = os.environ.get("PREDIBASE_GATEWAY", DEFAULT_API_ENDPOINT)

    if serving_endpoint is None:
        serving_endpoint = os.environ.get("PREDIBASE_SERVING_ENDPOINT", DEFAULT_SERVING_ENDPOINT)

    url = remove_trailing_slash(url)
    serving_endpoint = remove_trailing_slash(serving_endpoint)

    if token is None:
        token = os.environ.get("PREDIBASE_API_TOKEN")
        if token is None:
            raise ValueError("Set the PREDIBASE_API_TOKEN environment variable to use the Python SDK")

    session.url = url
    session.serving_http_endpoint = serving_endpoint
    session.serving_grpc_endpoint = serving_endpoint
    session.token = token

    if not token.startswith("pb_"):
        logger.warning(
            "The `PREDIBASE_API_TOKEN` long format you're using will be deprecated on April 15, 2024. "
            "Please upgrade your token by going to the Predibase UI and generating a new one.",
        )

    session.tenant = session.get_current_user()._tenant.short_code
    if verbose is not None:
        session.verbose = verbose


def get_session(gateway: str = None, token: str = None, serving_endpoint: str = None) -> Session:
    if _session.url is None:
        connect(url=gateway, serving_endpoint=serving_endpoint, token=token)

    # Return message upon successful client connection
    try:
        user = _session.get_current_user()

        if user is not None:
            log_info(f"Connected to Predibase as {user}")

    except Exception as e:
        raise ValueError(f"Unable to connect to Predibase: {e}")

    return _session


def set_verbose(value: bool):
    _session.verbose = value


def connect(
    url: Optional[str] = None,
    serving_endpoint: Optional[str] = None,
    token: Optional[str] = None,
    verbose: Optional[bool] = None,
):
    connect_session(_session, url, serving_endpoint, token, verbose)


def set_connection(source_id):
    if _session.url is None:
        connect()
    _session.set_connection(source_id)


def get_connections():
    if _session.url is None:
        connect()
    return _session.get_connections()


def execute(statement: str, params: Dict[str, Any] = None) -> pd.DataFrame:
    if _session.url is None:
        connect()
    return _session.execute(statement, params)



import concurrent.futures
from abc import abstractmethod
from typing import Dict, List, Optional, Union

import deprecation
import pandas as pd
from tabulate import tabulate

from predibase.pql.api import Session
from predibase.resource.connection import Connection
from predibase.resource.dataset import Dataset
from predibase.resource.llm import interface
from predibase.resource.llm.prompt import PromptTemplate
from predibase.version import __version__


class LlmMixin:
    session: Session

    def LLM(self, uri: str) -> Union["interface.HuggingFaceLLM", "interface.LLMDeployment"]:
        if uri.startswith("pb://deployments/"):
            return interface.LLMDeployment(self.session, uri[len("pb://deployments/") :])

        if uri.startswith("hf://"):
            return interface.HuggingFaceLLM(self.session, uri[len("hf://") :])

        raise ValueError(
            "must provide either a Hugging Face URI (hf://<...>) "
            "or a Predibase deployments URI (pb://deployments/<name>).",
        )

    def prompt_template(self, template_str: str) -> PromptTemplate:
        return PromptTemplate(template_str)

    def get_supported_llms(self):
        """Returns a list of supported HuggingFace LLMs."""
        data = self.session.get_json("/supported_llms")
        return data

    _simple_table_keys = ["name", "modelName", "deploymentStatus", "isShared"]

    def _rename_table_keys(self, key: str):
        return "deploymentName" if key == "name" else key

    def list_llm_deployments(self, active_only=False, print_as_table=True) -> Union[List[Dict], None]:
        """Fetches the current list of all LLM deployments regardless of status. Can either return this list or
        print it as a concise table. Can also optionally filter by active deployments.

        Returns: List of LLMDeployment objects or None
        """
        llms: List[Dict] = self.session.get_json(f'/llms?activeOnly={"true" if active_only else "false"}')
        if print_as_table:
            compressed_llms = [{self._rename_table_keys(k): llm[k] for k in self._simple_table_keys} for llm in llms]
            print(tabulate(compressed_llms, headers="keys", tablefmt="fancy_grid"))
        else:
            return llms

    @abstractmethod
    def get_dataset(self) -> Dataset:
        pass

    @abstractmethod
    def list_connections(self) -> List[Connection]:
        pass

    # TODO: remove this... again.
    @deprecation.deprecated(
        deprecated_in="2023.5.8",  # TODO: what was meant to be the date here
        current_version=__version__,
        details="Use the LLMDeployment().prompt method instead",
    )
    def prompt(
        self,
        templates: Union[str, List[str]],
        deployment_names: Union[str, List[str]],
        index: Optional[Union[str, Dataset]] = None,
        dataset: Optional[Union[str, Dataset]] = None,
        limit: Optional[int] = 10,
        options: Optional[Dict[str, float]] = None,
    ) -> pd.DataFrame:
        if not (index or dataset):
            # talk directly to the LLM, bypassing Temporal and the engines.
            return self.generate(templates=templates, deployment_names=deployment_names, options=options)

        # Set the connection if we have a dataset or index
        conn_id = None
        if isinstance(dataset, Dataset):
            conn_id = dataset.connection_id
        elif isinstance(index, Dataset):
            conn_id = index.connection_id
        resp = self.session.post_json(
            "/prompt",
            json={
                "connectionID": conn_id,
                "deploymentNames": [deployment_names] if isinstance(deployment_names, str) else deployment_names,
                "templates": [templates] if isinstance(templates, str) else templates,
                "options": options,
                "limit": limit,
                "indexName": index.name if isinstance(index, Dataset) else index,
                "datasetName": dataset.name if isinstance(dataset, Dataset) else dataset,
            },
            timeout=300,  # increase timeout to 5 minutes for this request
        )
        responses = [interface.GeneratedResponse(**r) for r in resp["responses"]]
        return interface.GeneratedResponse.to_pandas(responses)

    # TODO: remove this... again.
    @deprecation.deprecated(
        deprecated_in="2023.5.8",  # TODO: what was meant to be the date here
        current_version=__version__,
        details="Use the LLMDeployment().prompt method instead",
    )
    def generate(
        self,
        templates: Union[str, List[str]],
        deployment_names: Union[str, List[str]],
        options: Optional[Dict[str, float]] = None,
    ) -> pd.DataFrame:
        deployment_names = [deployment_names] if isinstance(deployment_names, str) else deployment_names
        templates = [templates] if isinstance(templates, str) else templates

        resp_list, future_to_args = [], dict()
        with concurrent.futures.ThreadPoolExecutor() as executor:
            futures = []
            for input_text in templates:
                for deployment_name in deployment_names:
                    endpoint = f"/{self.session.tenant}/deployments/v2/llms/{deployment_name}/generate"
                    post_func = self.session.post_json_serving
                    future = executor.submit(
                        post_func,
                        endpoint,
                        {
                            "inputs": input_text,
                            "parameters": options,
                        },
                    )
                    future_to_args[future] = (deployment_name, input_text)
                    futures.append(future)

            for future in concurrent.futures.as_completed(futures):
                try:
                    deployment_name, input_text = future_to_args[future]
                    res = future.result()
                    res = interface.GeneratedResponse(
                        prompt=input_text,
                        response=res["generated_text"],
                        model_name=deployment_name,
                        generated_tokens=res["details"]["generated_tokens"],
                    )
                    resp_list.append(res)
                except Exception as exc:
                    print("ERROR:", exc)
        return interface.GeneratedResponse.to_pandas(resp_list)



#! /usr/bin/env python
# Copyright (c) 2021 Predibase, Inc.
import datetime
import json
import logging
import os
import platform
import sys
import time
import urllib.request
from dataclasses import dataclass
from pathlib import Path
from threading import Event, Thread
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import pandas as pd
import requests
import requests.exceptions
import semantic_version
import websockets
from dateutil import parser
from dateutil.parser import ParserError
from decorator import decorator
from progress_table import ProgressTable
from tqdm import tqdm
from urllib3.util import Retry
from websockets.sync.client import connect

from predibase.pql.adapter import TimeoutHTTPAdapter
from predibase.pql.utils import get_results_df, retry
from predibase.resource.llm.util import print_events
from predibase.resource.user import User
from predibase.util import get_trace_id, log_error, log_warning
from predibase.util.json import JSONType
from predibase.util.metrics import (
    formatted_time_delta,
    get_metrics_table,
    is_llm_model,
    metricsRegex,
    model_logs_directory,
)
from predibase.version import __version__


@decorator
def warn_outdated_sdk(fn, *args, **kwargs):
    resp = fn(*args, **kwargs)
    server_release_version = resp.headers.get("X-Predibase-Release-Version", None)
    if server_release_version is not None:
        if server_release_version == "" or server_release_version == "staging":
            if not __version__.startswith("0.1.1+dev") and not __version__.startswith("2999"):
                log_warning(
                    "WARN: Using a post-release / prod version of the SDK in staging can lead to unexpected "
                    "behavior. Consider installing from latest master.",
                )
        else:
            server_release_semver = semantic_version.Version(server_release_version)
            sdk_semver = semantic_version.Version(__version__)
            if sdk_semver < server_release_semver:
                log_warning(
                    f"WARN: Currently installed SDK is outdated. This can lead to bugs or unexpected behavior. "
                    f"Consider upgrading to the latest version. Installed: {__version__} Latest: "
                    f"{server_release_version}.",
                )

    return resp


class PQLException(RuntimeError):
    def __init__(self, message):
        super().__init__(message)
        self.message = message


class ServerResponseError(RuntimeError):
    def __init__(self, message, code):
        super().__init__(message)
        self.message = message
        self.code = code


@dataclass
class Session:
    connection_id: int = None
    verbose: bool = False
    url: str = None
    serving_grpc_endpoint: str = None
    serving_http_endpoint: str = None
    tenant: str = "default"
    token: str = None
    timeout_in_seconds: int = 120

    def __post_init__(self):
        self._token_user_cache: Optional[Tuple[str, User]] = None

    def execute(
        self,
        statement: str,
        params: Dict[str, Any] = None,
        connection_id: Optional[int] = None,
        engine_id: Optional[int] = None,
    ) -> pd.DataFrame:
        if not self.is_plan_expired():
            if not statement.endswith(";"):
                statement += ";"

            params = params or {}
            if self.verbose:
                logging.info("-- EXECUTE:")
                logging.info(statement)

            conn = connection_id or self.connection_id
            resp = self._post(
                "/queries",
                json={
                    "connectionID": conn,
                    "rawQuery": statement,
                    "queryParams": params,
                    "engineID": engine_id,
                },
            )
            data = _to_json(resp)

            start_t = time.time()
            last_t = start_t

            query_status = data.get("queryStatus")
            if query_status:
                query_id = query_status["id"]
                while not query_status.get("completed"):
                    time.sleep(0.25)
                    query_status = self.get_query_status(query_id)
                    if self.verbose:
                        last_t = show_progress(last_t)

                status = query_status.get("status")
                if status == "completed":
                    if self.verbose:
                        logging.info(f"\n-- DONE: {round(time.time() - start_t, 2)}s")
                    # TODO: do not fetch results all at once, just return metadata like rowcount
                    return self.get_results(query_id)
                else:
                    raise PQLException(
                        f"Query completed with invalid status: {status}\n"
                        f"Query:\n{query_status['rawQuery']}\n"
                        f"Error:\n{query_status['errorText']}\n",
                    )

            raise PQLException(f"Invalid response: {data}")
        else:
            raise PermissionError(
                "Queries are locked for expired plans. Contact us to upgrade.",
            )

    @property
    def user(self) -> User:
        if self._token_user_cache is None or self._token_user_cache[0] != self.token:
            self._token_user_cache = (self.token, self.get_current_user())
        return self._token_user_cache[1]

    def get_current_user(self):
        resp = self.get_json("/users/current")
        return User.from_dict({"session": self, **resp})

    def is_free_trial(self):
        return self.user.tenant.subscription.tier == "free"

    def is_plan_expired(self):
        days_remaining = self.user.tenant.subscription.days_remaining_in_plan

        if days_remaining is None:
            return False

        return days_remaining <= 0

    @retry(times=20, exceptions=(requests.exceptions.ChunkedEncodingError,))
    def get_query_status(self, query_id: int) -> str:
        resp = self._get(f"/queries/{query_id}")
        data = _to_json(resp)
        return data["queryStatus"]

    def get_results(self, query_id: int) -> pd.DataFrame:
        retry_backoff = 10  # seconds
        max_retries = 180  # 6 retries per minute, * 30 minutes
        current_retry_count = 0

        while True:
            resp = self._get(f"/queries/{query_id}/results")
            data = _to_json(resp)

            if resp.status_code == 202:
                # This means that there was an issue w/ the request, engine status was not active
                if current_retry_count >= max_retries:
                    raise PQLException(f'Unable to fetch query results, engine status: {data["engineStatus"]}')

                current_retry_count += 1
                time.sleep(retry_backoff)
                continue

            return get_results_df(data)

    def get_dataset_id(self, name: str) -> int:
        datasets = self.get_datasets()
        matched_datasets = datasets.loc[datasets["name"] == name]
        return int(matched_datasets.iloc[-1].id)

    def set_connection(self, connection_id_or_name: Union[str, int]):
        self.connection_id = self.get_connection_id(connection_id_or_name)

    def get_connections(self) -> pd.DataFrame:
        resp = self._get("/connections")
        data = _to_json(resp).get("connections")
        if data is None:
            return pd.DataFrame()
        return pd.DataFrame.from_dict(data)

    def get_connection_id(self, connection_id_or_name: Union[str, int]):
        if isinstance(connection_id_or_name, str):
            endpoint = f"/connections/name/{connection_id_or_name}"
            resp = self.get_json(endpoint)
            return int(resp["id"])
        return connection_id_or_name

    def get_datasets(self) -> pd.DataFrame:
        resp = self._get("/data/datasets/fetch/allowed_only")
        data = _to_json(resp).get("datasets")
        if data is None:
            return pd.DataFrame()
        return pd.DataFrame.from_dict(data)

    def head(self, endpoint) -> requests.Response:
        return self._http().head(self.url + endpoint, headers=self._get_headers())

    def post(self, endpoint, data: Any = None, json: Any = None, **kwargs):
        resp = self._post(endpoint, data=data, json=json, **kwargs)
        return _to_json(resp)

    def post_json(self, endpoint: str, json: Any, **kwargs):
        resp = self._post(endpoint, json=json, **kwargs)
        return _to_json(resp)

    def post_json_serving(self, endpoint: str, json: Any, **kwargs):
        resp = self._post_serving(endpoint, json=json, **kwargs)
        return _to_json(resp)

    def get_json_serving(self, endpoint: str, params: dict = None):
        resp = self._get_serving(endpoint, params=params)
        return _to_json(resp)

    def put_json(self, endpoint: str, json: Any):
        resp = self._put(endpoint, json)
        return _to_json(resp)

    def delete_json(self, endpoint: str):
        res = self._delete(endpoint)
        # Handle errors / print out the trace
        return _to_json(res)

    def get_json_until(
        self,
        endpoint: str,
        success_cond: Callable,
        error_cond: Optional[Callable] = None,
    ):
        while True:
            resp = self.get_json(endpoint)
            if success_cond(resp):
                return resp
            if error_cond is not None:
                base_err = error_cond(resp)
                if base_err:
                    err_text = resp.get("modelVersion", {}).get("errorText", "Unknown")
                    error_msg = f"{base_err}\n\n" f"Remote error:\n\n{err_text}\n"
                    raise RuntimeError(error_msg)

            time.sleep(1.0)

    def get_model_logs_urls(
        self,
        model_id: int,
    ) -> List[Dict[str, str]]:
        """Get the logs associated with this model for all runs.

        Args:
            model_id: The model ID.

        Returns:
            List of a tuple Ludwig Log Files and a presigned URL to download them.
            [{
                "filepath": "ludwig_logs/2021-05-27T18:50:13.000Z/2021-05-27T18:50:13.000Z.log",
                "url": "https://s3.amazonaws.com/2021-05-27T18:50:13.000Z/2021-05-27T18:50:13.000Z.log?AWSAccessKeyId=AKIAJ2...", # noqa E501
            }]
        """
        try:
            resp = self.get_json(
                f"/models/version/{model_id}/logs",
            )

            return resp
        except ServerResponseError as e:
            raise e

    def get_model_logs(
        self,
        model_id: int,
        sync_event: Event,
        active_run_id: Optional[str] = None,
        parent_dir: [str] = "",
    ) -> None:
        """Sync the logs associated with this model for all runs.

        Args:
            model_id: The model ID.
            sync_event: The event to use to indicate when to stop syncing logs between the client and blob storage.
            active_run_id: The run ID to sync. If None, sync all runs.
            parent_dir: The parent directory to sync the logs to. If None, sync relative to where the command is run.
        """

        def _download_logs(logfiles):
            for logfile in logfiles:
                log_file_path = os.path.join(parent_dir, logs_dir_name, logfile["filepath"])
                Path(os.path.dirname(log_file_path)).mkdir(parents=True, exist_ok=True)
                try:
                    urllib.request.urlretrieve(logfile["url"], log_file_path)
                except Exception:
                    # TODO: is there a way to safely print errors without breaking the progress table?
                    # print(f"Failed to sync {logfile['url']}: {e}")
                    pass

        logfiles = []
        moreLogsAvailable = True
        logs_dir_name = model_logs_directory(model_id)
        Path(logs_dir_name).mkdir(exist_ok=True)
        # Shorten Active Run ID to match UI and response from API
        if active_run_id is not None:
            active_run_id = active_run_id[:6]

        # When a model moves into training, there can be a considerable delay before the logs are available.
        # We need to wait for the logs to be available before we can sync them.
        # This should occur after the first batch/step has finished.
        while moreLogsAvailable and not sync_event.is_set():
            logs = self.get_model_logs_urls(model_id)
            logfiles = logs.get("logfiles", [])
            moreLogsAvailable = logs.get("moreLogsAvailable", False)

            if active_run_id is not None:
                # Filter the list of logfiles to only include the active run
                logfiles = [logfile for logfile in logfiles if f"{active_run_id}/" in logfile["filepath"]]

            if len(logfiles) > 0:
                _download_logs(logfiles)

            # The model is in a terminal state, so we can stop syncing logs
            if not moreLogsAvailable:
                break

            time.sleep(30)

    def _get_elapsed(self, curr_status: str, timeline_resp: Dict[str, Any]):
        start_time = None
        try:
            created = timeline_resp.get(curr_status, {}).get("created", None)
            start_time = parser.parse(created)
        except ParserError:
            start_time = None

        if start_time is None:
            return "{elapsed}"

        elapsed = datetime.datetime.utcnow() - start_time.replace(tzinfo=None)
        elapsed = elapsed - datetime.timedelta(microseconds=elapsed.microseconds)

        return str(elapsed)

    def _wait_until_model_ready_with_logging(
        self,
        model_id: int,
        launch_tensorboard: bool,
    ):
        launch_tb_parameter = f"{launch_tensorboard}".lower()
        runs_endpoint = f"/models/version/{model_id}?withRuns={launch_tb_parameter}"
        timeline_endpoint = f"/models/version/{model_id}/timeline"

        success_cond = lambda resp: (
            resp["modelVersion"]["status"] == "ready"
            or resp["modelVersion"]["status"] == "deploying"
            or resp["modelVersion"]["status"] == "deployed"
            or resp["modelVersion"]["status"] == "undeploying"
            or resp["modelVersion"]["status"] == "canceled"
        )
        error_cond = lambda resp: (
            (
                f"Failed to train model with status {resp['modelVersion']['status'].upper()} and error "
                f"{resp['modelVersion']['errorText']}"
            )
            if resp["modelVersion"]["status"] in {"failed"}
            else None
        )

        last_status = None

        pbars = []
        training_pbar = None
        metrics_thread = None
        tensorboard_thread = None
        tensorboard_event = Event()

        def update_last_pbar(last_status: str, timeline_resp: Dict[str, Any]):
            if pbars:
                pbars[-1].bar_format = "✓ " + pbar.desc + f" {self._get_elapsed(last_status, timeline_resp)}"
                pbars[-1].close()

        while True:
            resp = self.get_json(runs_endpoint)
            timeline_resp = self.get_json(timeline_endpoint)

            if success_cond(resp):
                curr_status = resp["modelVersion"]["status"]
                update_last_pbar(last_status, timeline_resp)

                curr_status = curr_status.capitalize()
                pbar = tqdm(
                    None,
                    bar_format=curr_status.capitalize(),
                    desc=curr_status.capitalize(),
                    ncols=0,
                    miniters=1,
                )
                pbar.update()
                pbar.close()
                return resp
            if error_cond is not None:
                base_err = error_cond(resp)
                if base_err:
                    curr_status = resp["modelVersion"]["status"].capitalize()
                    pbar = tqdm(
                        None,
                        bar_format=curr_status.capitalize(),
                        desc=curr_status.capitalize(),
                        ncols=0,
                    )
                    pbar.update()
                    pbar.close()

                    err_text = resp.get("modelVersion", {}).get("errorText", "Unknown")
                    error_msg = f"{base_err}\n\n" f"Remote error:\n\n{err_text}\n"
                    raise RuntimeError(error_msg)
            curr_status = resp["modelVersion"]["status"]
            config = resp["modelVersion"]["config"]
            if curr_status != last_status:
                update_last_pbar(last_status, timeline_resp)
                last_status = curr_status
                if curr_status != "training":
                    pbar = tqdm(
                        None,
                        bar_format="  "
                        + curr_status.capitalize()
                        + f"... {self._get_elapsed(curr_status, timeline_resp)}",
                        desc=curr_status.capitalize(),
                        ncols=0,
                    )
                    pbars.append(pbar)
                elif not metrics_thread:
                    training_pbar = tqdm(
                        None,
                        bar_format=f"  Setting up training run... {self._get_elapsed(curr_status, timeline_resp)}",
                        desc=curr_status.capitalize(),
                        ncols=0,
                        leave=False,
                    )

                    start_time = datetime.datetime.now()
                    model_id = resp["modelVersion"]["id"]
                    metrics_thread = Thread(
                        target=self._stream_model_metrics,
                        args=(model_id, config, get_metrics_table(), training_pbar, start_time, 10),
                        daemon=True,
                    )
                    metrics_thread.start()

                # Stop threads if we're no longer training
                if tensorboard_thread and curr_status != "training":
                    tensorboard_event.set()
                    tensorboard_thread.join()
                    tensorboard_thread = None

            # Start a thread to sync the logs after TensorBoard so it's not blocking
            # We have to wait until the model is both training and has an active run
            if curr_status == "training":
                # Runs is not populated until the pre-training calcs are complete
                run_id = None
                runs = resp["runs"]
                if runs is not None and len(runs) > 0:
                    run_id = runs[0].get("info", {}).get("run_id", None)
                if tensorboard_thread is None and run_id is not None:
                    tensorboard_thread = Thread(
                        target=self.get_model_logs,
                        args=[model_id, tensorboard_event, run_id],
                        daemon=True,
                    )
                    tensorboard_thread.start()
            if pbars:
                pbars[-1].bar_format = (
                    "  " + curr_status.capitalize() + f"... {self._get_elapsed(curr_status, timeline_resp)}"
                )
                pbars[-1].refresh()
                pbars[-1].update()
            if training_pbar is not None and not training_pbar.disable:
                training_pbar.bar_format = (
                    f"  Setting up training run... {self._get_elapsed(curr_status, timeline_resp)}"
                )
                training_pbar.refresh()
                training_pbar.update()

            time.sleep(1.0)

    def get_llm_deployment_events_until_with_logging(
        self,
        events_endpoint: str,
        success_cond: Callable,
        error_cond: Optional[Callable] = None,
    ):
        print_header = True
        existing_events = []
        while True:
            events = self.get_json(events_endpoint)
            print_events(events=events, print_header=print_header, existing_events=existing_events)
            print_header = False
            existing_events = [(event["eventType"], event["timestamp"]) for event in events.get("ComputeEvents", [])]
            if success_cond(events):
                return
            if error_cond(events):
                return
            time.sleep(2)
        pass

    def get_llm_deployment_until_with_logging(
        self,
        endpoint: str,
        success_cond: Callable,
        error_cond: Optional[Callable] = None,
    ):
        last_status = None
        pbars = []

        while True:
            resp = self.get_json(endpoint)
            if success_cond(resp):
                if pbars:
                    pbars[-1].close()
                curr_status = resp["deploymentStatus"].capitalize()
                pbar = tqdm(None, bar_format=curr_status.capitalize(), desc=curr_status.capitalize(), ncols=0)
                pbar.update()
                pbar.close()
                return resp
            if error_cond is not None:
                base_err = error_cond(resp)
                if base_err:
                    curr_status = resp["deploymentStatus"].capitalize()
                    pbar = tqdm(None, bar_format=curr_status.capitalize(), desc=curr_status.capitalize(), ncols=0)
                    pbar.update()
                    pbar.close()

                    # TODO(hungcs): Add back when we get deployment errors
                    # err_text = resp.get("modelVersion", {}).get("errorText", "Unknown")
                    # error_msg = f"{base_err}\n\n" f"Remote error:\n\n{err_text}\n"
                    raise RuntimeError(base_err)
            curr_status = resp["deploymentStatus"]
            if curr_status != last_status:
                if pbars:
                    pbar = pbars[-1]
                    pbar.bar_format = pbar.desc + " {elapsed}"
                    pbar.close()
                last_status = curr_status
                pbar = tqdm(
                    None,
                    bar_format=curr_status.capitalize() + "... {elapsed}",
                    desc=curr_status.capitalize(),
                    ncols=0,
                )
                pbars.append(pbar)
            if pbars:
                pbars[-1].update()
            time.sleep(1.0)

    def wait_for_dataset(self, endpoint, until_fully_connected: bool = False):
        while True:
            resp = self.get_json(endpoint)
            if "lastError" in resp and resp["lastError"]:
                raise ValueError("Error waiting for dataset", resp["lastError"])
            if until_fully_connected:
                if resp["status"] == "connected":
                    return resp
            else:
                return resp
            time.sleep(1.0)

    def get_json(self, endpoint: str, params: dict = None):
        resp = self._get(endpoint, params=params)
        return _to_json(resp)

    def get_websocket(self, endpoint: str) -> websockets.sync.client.ClientConnection:
        ws_url = self.url.replace("https://", "wss://").replace("http://", "ws://")
        return connect(ws_url + endpoint, additional_headers=self._get_headers())

    def _http(self):
        """Returns a http requests configured with back off and timeout. Will timeout after 20 seconds, and will
        sleep for time based on:

        {backoff factor} * (2 ** ({number of total retries} - 1))
        """
        # see: https://findwork.dev/blog/advanced-usage-python-requests-timeouts-retries-hooks/
        retry_strategy = Retry(
            total=4,
            backoff_factor=2,
            status_forcelist=[104, 429, 500, 502, 503, 504],  # Exclude bad request
            allowed_methods=["HEAD", "GET", "PUT", "PATCH", "POST", "OPTIONS"],
        )
        http = requests.Session()
        adapter = TimeoutHTTPAdapter(
            max_retries=retry_strategy,
            timeout=self.timeout_in_seconds,
        )
        http.mount("https://", adapter)
        http.mount("http://", adapter)
        return http

    # def _ws(self):
    #     http = requests.Session()
    #     http.mount("wss://", adapter)
    #     http.mount("ws://", adapter)
    #     return http

    def _get_headers(self):
        return {
            "Authorization": "Bearer " + self.token,
            "User-Agent": f"predibase-sdk/{__version__} ({platform.version()})",
        }

    @warn_outdated_sdk
    def _post_serving(self, endpoint: str, data: Any = None, json: Any = None, **kwargs):
        return self._http().post(
            "https://" + self.serving_http_endpoint + endpoint,
            data=data,
            json=json,
            headers=self._get_headers(),
            **kwargs,
        )

    @warn_outdated_sdk
    def _get_serving(self, endpoint: str, params: dict = None):
        return self._http().get(
            "https://" + self.serving_http_endpoint + endpoint,
            headers=self._get_headers(),
            timeout=(self.timeout_in_seconds, self.timeout_in_seconds),
        )

    @warn_outdated_sdk
    def _post(self, endpoint: str, data: Any = None, json: Any = None, **kwargs):
        return self._http().post(self.url + endpoint, data=data, json=json, headers=self._get_headers(), **kwargs)

    @warn_outdated_sdk
    def _get(self, endpoint: str, params: dict = None):
        return self._http().get(
            self.url + endpoint,
            headers=self._get_headers(),
            params=params,
            timeout=(self.timeout_in_seconds, self.timeout_in_seconds),
        )

    @warn_outdated_sdk
    def _get_ws(self, endpoint: str):
        ws_url = self.url.replace("https://", "wss://").replace("http://", "ws://")
        return requests.Session().get(ws_url + endpoint, headers=self._get_headers(), timeout=None, stream=True)

    @warn_outdated_sdk
    def _delete(self, endpoint: str):
        return self._http().delete(
            self.url + endpoint,
            headers=self._get_headers(),
            timeout=(self.timeout_in_seconds, self.timeout_in_seconds),
        )

    @warn_outdated_sdk
    def _put(self, endpoint: str, json_data: Any):
        return self._http().put(self.url + endpoint, json=json_data, headers=self._get_headers())

    def _stream_model_metrics(
        self,
        model_id: str,
        config: Dict,
        table: ProgressTable,
        training_pbar: tqdm,
        start_time: datetime,
        max_attempts: int = 10,
    ):
        ws = None
        ws_data = []
        attempts = 0

        def print_progress_bar(resp: Dict):
            if "steps" in resp["meta"] and resp["meta"]["steps"] > 0:
                table.progress_bar_active = True
                total_steps = resp["meta"]["total_steps"]
                steps = resp["meta"]["steps"]

                table._print_progress_bar(
                    steps,
                    total_steps,
                    show_before=f" {steps}/{total_steps} steps ",
                )

        while True:
            try:
                if ws is None:
                    attempts += 1
                    ws = self.get_websocket(f"/models/metrics/history/stream/{model_id}")
                resp = ws.recv()
                resp = json.loads(resp)

                if "steps" in resp["meta"] and resp["meta"]["steps"] > 0:
                    if not training_pbar.disable:
                        training_pbar.close()
                        time.sleep(0.5)  # stops race conditions with steps progress bar

                    if not table.header_printed:
                        table._print_header(top=True)

                print_progress_bar(resp)
                metrics_data = resp["data"]
                if len(metrics_data) > 0:
                    table.progress_bar_active = False
                    table._print_row()

                    ws_data.append(metrics_data)
                    run_ids = metrics_data.keys()
                    if resp["meta"]["is_hyperopt"]:
                        run_id = max(run_ids, key=lambda rid: metrics_data[rid][-1]["epoch"])
                    else:
                        run_id = next(iter(run_ids))

                    table.progress_bar_active = False
                    for epoch_data in metrics_data[run_id]:
                        # table.next_row()
                        epoch_in_data = epoch_data["epoch"]
                        table["epochs"] = epoch_in_data
                        table["time"] = formatted_time_delta(datetime.datetime.now(), start_time)

                        filtered_epoch_data = {
                            k: epoch_data[k]
                            for k in epoch_data
                            if k.startswith(("train_metrics", "validation_metrics", "test_metrics"))
                        }

                        # https://predibase.slack.com/archives/C04L07442JU/p1695772621939679?thread_ts=1695750475.580479&cid=C04L07442JU
                        if is_llm_model(config):
                            llm_filtered_epoch_data = {}
                            metrics_to_keep = ["loss", "next_token_perplexity", "perplexity", "word_error_rate", "bleu"]
                            for k in filtered_epoch_data:
                                match = metricsRegex.match(k)
                                if match:
                                    metric_name = match[3]
                                    if any(m == metric_name for m in metrics_to_keep):
                                        llm_filtered_epoch_data[k] = filtered_epoch_data[k]
                            filtered_epoch_data = llm_filtered_epoch_data

                        metrics_dict = {}
                        for full_metric_name in filtered_epoch_data.keys():
                            match = metricsRegex.match(full_metric_name)
                            if match:
                                split, feature, metric_name = match[1], match[2], match[3]
                                if feature not in metrics_dict:
                                    metrics_dict[feature] = {}
                                if metric_name not in metrics_dict[feature]:
                                    metrics_dict[feature][metric_name] = {}
                                metrics_dict[feature][metric_name][split] = filtered_epoch_data[full_metric_name]
                        last_feature = None
                        features = sorted(
                            metrics_dict.keys(),
                            key=lambda x: metrics_to_keep.index(x) if x in metrics_to_keep else 100,
                        )
                        for i, feature in enumerate(features):
                            for j, metric_name in enumerate(metrics_dict[feature]):
                                if feature != last_feature:
                                    last_feature = feature
                                    table["feature"] = feature
                                table["metric"] = metric_name
                                table["train"] = metrics_dict[feature][metric_name]["train_metrics"]
                                if "validation_metrics" in metrics_dict[feature][metric_name]:
                                    table["val"] = metrics_dict[feature][metric_name]["validation_metrics"]
                                table["test"] = metrics_dict[feature][metric_name]["test_metrics"]

                                is_final_row_in_epoch = i == len(features) - 1 and j == len(metrics_dict[feature]) - 1
                                table.next_row(split=is_final_row_in_epoch)
                    table.progress_bar_active = True
                    time.sleep(0.5)  # stops race conditions with steps progress bar
                if resp["meta"]["is_completed"]:
                    print_progress_bar(resp)
                    # Close the websocket and table connections before returning.
                    table.close()
                    ws.close()
                    return

            except websockets.exceptions.ConnectionClosedError:
                if 0 < max_attempts <= attempts:
                    print(
                        f"Model metrics streaming failed due to a connection error after {max_attempts} attempts. The "
                        f"training run is still ongoing and is NOT affected. You can monitor progress in the UI.",
                    )
                    # Close the websocket and table connections before returning.
                    table.close()
                    ws.close()
                    return

                # If we still have attempts remaining, reset the websocket object and keep going
                ws = None
                continue

            except websockets.exceptions.ConnectionClosedOK:
                return

            # TODO (hungcs): Handle specific error types
            except Exception:
                continue


def show_progress(last_t):
    t = time.time()
    if t - last_t > 1:
        sys.stdout.write("")
        sys.stdout.flush()
        return t
    return last_t


def _to_json(resp: requests.Response) -> JSONType:
    if resp.status_code != 200:
        if resp.status_code == 202:
            # Processing in progress
            return {}
        trace_id = get_trace_id(resp)
        raise ServerResponseError(
            f"Error {resp.status_code}: {_get_error(resp)}. Trace ID: {trace_id}",
            resp.status_code,
        )
    if resp.content:
        try:
            data = resp.json()
            if data:
                if type(data) is dict:
                    error_message = data.get("errorMessage")
                    if error_message:
                        # TODO: very strange to call these PQLExceptions.
                        raise PQLException(error_message)
            return data
        except requests.exceptions.JSONDecodeError:
            log_error(f"Failed to decode payload as JSON. Payload text: \n{resp.text}\n")
            raise

    return {}


def _get_error(resp):
    try:
        data = resp.json()
        if data is None:
            return "Unknown server error"
        if "error" in data:
            return data["error"]
        return resp.reason
    except requests.exceptions.JSONDecodeError:
        return f"Failed to decode payload as JSON. Payload text: \n{resp.text}\n"



from __future__ import annotations

import os
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Optional, Dict, Tuple

import yaml
from tabulate import tabulate

from predibase._errors import PredibaseError
from predibase.resource.llm.prompt import PromptTemplate
from predibase.resources.model import PretrainedHuggingFaceLLM, FinetunedLLMAdapter
from predibase.util import log_info, get_url

if TYPE_CHECKING:
    from predibase import Predibase
    from predibase.resources.dataset import Dataset
    from predibase.resources.repo import Repo


_PATH_HERE = os.path.abspath(os.path.dirname(__file__))
_TEMPLATE_DIR = os.path.join(_PATH_HERE, "../resource/llm/templates")
_CONFIG_FILENAME = "config.yaml"  # Name of config file for loading templates.


class Adapters:
    def __init__(self, client: Predibase):
        self._client = client
        self._session = client._session  # Directly using the session in the short term as we transition to v2.

    def create(
        self,
        config: FinetuneConfig,
        *,
        description: str | None = None,
        repo: str | Repo | None = None,
        wait: bool = True,
        show_tensorboard: bool = True,
    ):
        # TODO(jeff): docstring

        if show_tensorboard and not wait:
            raise RuntimeError("`show_tensorboard` is a blocking option and thus requires `wait` to be True.")

        if repo is None:
            # If no repo is specified, automatically construct the repo name from the dataset and model name.
            dataset_name = config.dataset.name
            if "/" in dataset_name:
                _, dataset_name = dataset_name.split("/")

            model_name = config.base_model.name
            if "/" in model_name:
                _, model_name = model_name.split("/")

            repo = f"{model_name}-{dataset_name}"

        if isinstance(repo, str):
            if "/" in repo:
                repo = repo.replace("/", "-")
            repo = self._client.repos.create(repo, exists_ok=True)

        train_req = {
            "config": config.render(),
            "datasetID": config.dataset.id,
            "repoID": repo.id,
            "description": description,
        }

        resp = self._session.post_json("/models/train", train_req)
        model_id = resp["model"]["id"]

        # Output link to UI model page
        endpoint = f"models/version/{model_id}"
        url = get_url(self._session, endpoint)
        log_info(f"Check Status of Model Training Here: [link={url}]{url}")

        adapter = FinetunedLLMAdapter(self._client, repo.name, model_id)
        if wait:
            return adapter.wait_ready(show_tensorboard=show_tensorboard)
        return adapter

    @staticmethod
    def FinetuneConfig(*args, **kwargs) -> FinetuneConfig:
        return FinetuneConfig(*args, **kwargs)

    def get(self, adapter_uri: str) -> FinetunedLLMAdapter:
        if adapter_uri.startswith("pb://adapters/"):
            adapter_uri = adapter_uri[len("pb://adapters/") :]

        tokens = adapter_uri.split("/")
        if len(tokens) < 2:
            raise PredibaseError(f"expected a URI of the form [pb://adapters/]<repo_name>/<version>")

        resp = self._session.get_json(f"/models/version/name/{tokens[0]}?version={tokens[1]}")
        if not resp.get("llmBaseModelName", ""):
            raise PredibaseError(f"uri {adapter_uri} does not point to a fine-tuned LLM adapter")

        return FinetunedLLMAdapter(self._client, tokens[0], resp["id"])


class FinetuneConfig:
    def __init__(
        self,
        *,
        base_model: str | PretrainedHuggingFaceLLM,
        dataset: Dataset,  # TODO: enable passing dataset by name once we get rid of connection namespacing
        prompt_template: str | PromptTemplate,
        target: str,
        finetune_template: FinetuneTemplate | None = None,
        epochs: int | None = None,
        train_steps: int | None = None,
        learning_rate: float | None = None,
    ):

        if train_steps is not None and epochs is not None:
            raise ValueError("Cannot specify both train_steps and epochs.")

        self._base_model = base_model
        if isinstance(self._base_model, str):
            self._base_model = PretrainedHuggingFaceLLM(self._base_model)

        self._dataset = dataset
        self._prompt_template = prompt_template
        self._target = target
        self._finetune_template = finetune_template or FinetuneTemplateCollection().default
        self._epochs = epochs
        self._train_steps = train_steps
        self._learning_rate = learning_rate

    @property
    def base_model(self) -> PretrainedHuggingFaceLLM:
        return self._base_model

    @property
    def dataset(self) -> Dataset:
        return self._dataset

    def render(self) -> Dict:
        prompt_template = self._prompt_template
        if isinstance(prompt_template, PromptTemplate):
            prompt_template = prompt_template.render()

        config = self._finetune_template.to_config(self._base_model, prompt_template, self._target)

        # HACK(Arnav): Temporary hack to inject the right target_modules into the config for mixtral and
        # mixtral instruct. This should be removed once the Mixtral models have a default PEFT mapping for LoRA
        # target_modules. See MLX-1680: https://linear.app/predibase/issue/MLX-1680/remove-peftlora-mixtral-hack-in-the-predibase-sdk # noqa
        if (
            config.get("base_model", "") in {"mistralai/Mixtral-8x7B-v0.1", "mistralai/Mixtral-8x7B-Instruct-v0.1"}
            and "adapter" in config
            and config.get("adapter", {}).get("target_modules", None) is None
        ):
            config["adapter"]["target_modules"] = ["q_proj", "v_proj"]

        # Apply first-class training parameters.
        if self._train_steps is not None:
            config["trainer"]["train_steps"] = self._train_steps
        if self._epochs is not None:
            config["trainer"]["epochs"] = self._epochs
        if self._learning_rate is not None:
            config["trainer"]["learning_rate"] = self._learning_rate

        return config


class FinetuneTemplateCollection(dict):
    def __init__(self):
        super().__init__()

        for filename in os.listdir(_TEMPLATE_DIR):
            if not filename.endswith(".yaml"):
                continue

            if filename == _CONFIG_FILENAME:
                with open(os.path.join(_TEMPLATE_DIR, _CONFIG_FILENAME)) as config_file:
                    self._config = yaml.safe_load(config_file)
            else:
                tmpl = FinetuneTemplate(filename=os.path.join(_TEMPLATE_DIR, filename))
                self[tmpl.name] = tmpl

    @property
    def default(self) -> FinetuneTemplate:
        default_name = self._config.get("default", "")
        tmpl = self.get(default_name, None)
        if tmpl is None:
            raise RuntimeError(f"the default template '{default_name}' was not found")

        return tmpl

    def compare(self):
        print(self)

    def __str__(self):
        def make_row(key) -> Tuple[str, str, str]:
            tmpl = self[key]
            default_indicator = "------>" if tmpl.name == self._config["default"] else ""
            return default_indicator, key, tmpl.description

        return tabulate(
            (make_row(k) for k in sorted(self.keys())),
            tablefmt="simple_grid",
            headers=["Default", "Name", "Description"],
            colalign=["center"],
        )


@dataclass
class FinetuneTemplate:
    filename: str
    _meta: Optional[dict] = field(default=None, init=False)
    _raw_template_str: Optional[str] = field(default=None, init=False)

    @property
    def name(self) -> str:
        if self._meta is None:
            self._load()
        return self._meta["name"]

    @property
    def description(self) -> str:
        if self._meta is None:
            self._load()
        return self._meta["description"]

    def to_config(
        self,
        model: str | PretrainedHuggingFaceLLM,
        prompt_template: str = "__PROMPT_TEMPLATE_PLACEHOLDER__",
        target: str = "__TARGET_PLACEHOLDER__",
    ) -> dict:
        if self._raw_template_str is None:
            self._load()

        if isinstance(model, PretrainedHuggingFaceLLM):
            model = model.name

        cfg = yaml.safe_load(
            self._raw_template_str.format(
                base_model=model,
                prompt_template=prompt_template,
                target=target,
            ),
        )
        del cfg["template_meta"]
        return cfg

    def _load(self):
        with open(os.path.join(_TEMPLATE_DIR, self.filename)) as f:
            self._raw_template_str = f.read()

        self._meta = yaml.safe_load(self._raw_template_str)["template_meta"]



import os

from ._client import Predibase
from .client import PredibaseClient
from .predictor import AsyncPredictor, Predictor
from .version import __version__  # noqa: F401
from .resource.llm.prompt import PromptTemplate

import logging

logger = logging.getLogger(__name__)

__all__ = ["Predibase", "PredibaseClient", "AsyncPredictor", "Predictor", "PromptTemplate"]


def _configure_logging():
    log_level = os.getenv("PREDIBASE_LOG_LEVEL", "").lower()
    requests_logger = logging.getLogger("requests")
    urllib3_logger = logging.getLogger("urllib3")

    if not log_level:
        requests_logger.setLevel(logging.CRITICAL)
        urllib3_logger.setLevel(logging.CRITICAL)
    elif log_level == "debug":
        logger.setLevel(logging.DEBUG)
        requests_logger.setLevel(logging.DEBUG)
        urllib3_logger.setLevel(logging.DEBUG)
    elif log_level == "info":
        logger.setLevel(logging.INFO)
        requests_logger.setLevel(logging.INFO)
        urllib3_logger.setLevel(logging.INFO)


_configure_logging()



# CREATE_MODEL_TEMPLATE = """
#     CREATE MODEL {model_name} FROM {dataset}
#     CONFIG {config};
# """
from dataclasses import dataclass, field
from enum import Enum
from typing import Optional

import pandas as pd
from dataclasses_json import config, dataclass_json, LetterCase

from predibase.pql.api import Session
from predibase.resource.user import User


class QueryStatus(Enum):
    CREATED = "created"
    QUEUED = "queued"
    TRAINING = "training"
    RUNNING = "running"
    COMPLETED = "completed"
    CANCELED = "canceled"
    FAILED = "failed"

    def __str__(self):
        return str(self.value)

    @staticmethod
    def is_terminal_status(status: str) -> bool:
        return status in [QueryStatus.COMPLETED, QueryStatus.CANCELED, QueryStatus.FAILED]


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class Query:
    session: Session
    _id: int = field(metadata=config(field_name="id"))
    _raw_query: str = field(metadata=config(field_name="rawQuery"))
    _created: str = field(metadata=config(field_name="created"))
    _status: QueryStatus = field(metadata=config(field_name="status"))
    _results_url: Optional[str] = field(metadata=config(field_name="errorText"), default=None)
    _error_text: Optional[str] = field(metadata=config(field_name="errorText"), default=None)
    _created_by_user_id: Optional[int] = field(metadata=config(field_name="createdByUserID"), default=None)
    _status_text: Optional[str] = field(metadata=config(field_name="statusText"), default=None)
    _connection_id: Optional[int] = field(metadata=config(field_name="connectionID"), default=None)
    _completed: Optional[str] = field(metadata=config(field_name="completed"), default=None)
    _user: Optional[User] = field(metadata=config(field_name="user"), default=None)

    def __repr__(self):
        return (
            f"Query(id={self.id}, raw_query={self._raw_query}, status={self.status}, author={self.author}, "
            f"results_url={{...}}, error_text={self._error_text}, "
            f"created={self._created}, completed={self.completed})"
        )

    @property
    def id(self):
        return self._id

    @property
    def raw_query(self):
        return self._raw_query

    @property
    def created(self):
        return self._created

    @property
    def completed(self):
        return self._completed

    @property
    def author(self):
        return self._user.username if self._user else None

    @property
    def status(self):
        if QueryStatus.is_terminal_status(self._status):
            return self._status
        return self._get()["status"]

    def _get(self):
        return self.session.get_json(f"/queries/{self.id}")["queryStatus"]

    """Returns a link to a gzipped CSV file containing the results of the query."""

    def export_results(self) -> str:
        resp = self.session.get_json(f"queries/{self.id}/results/exported")
        return resp

    def get_results(self) -> pd.DataFrame:
        resp = self.session.get_json(f"/queries/{self.id}/results")
        if "errorMessage" in resp:
            raise RuntimeError(resp["errorMessage"])

        dataset = resp["dataset"]
        if "data" not in dataset:
            return pd.DataFrame()
        return pd.DataFrame(
            data=dataset["data"],
            columns=dataset["columns"],
            index=dataset.get("index"),
        )

    def cancel(self):
        self.session.post_json(f"/queries/{self.id}/cancel", {})



import logging
import os
import platform
from typing import Optional, Dict, Any
from urllib.parse import urljoin

import requests

from predibase._errors import PredibaseError, warn_outdated_sdk
from predibase.pql import start_session
from predibase.resource.user import User
from predibase.resources.adapters import Adapters
from predibase.resources.completions import Completions
from predibase.resources.datasets import Datasets
from predibase.resources.deployments import Deployments
from predibase.resources.models import Models
from predibase.resources.repos import Repos
from predibase.util.settings import load_settings
from predibase.version import __version__

_DEFAULT_API_GATEWAY = "https://api.app.predibase.com/v1"
_DEFAULT_SERVING_ENDPOINT = "serving.app.predibase.com"


logger = logging.getLogger(__name__)


class Predibase:
    def __init__(
        self,
        *,
        api_token: Optional[str] = None,
    ):
        """
        Create a new Predibase client instance.

        If not provided, optional params will be set by, in order of precedence:
        * Current environment variables
        * Hard-coded defaults

        :param api_token: The authentication token for the Predibase API.
        """
        api_token = api_token or os.environ.get("PREDIBASE_API_TOKEN") or load_settings().get("token")
        api_gateway = os.environ.get("PREDIBASE_GATEWAY") or _DEFAULT_API_GATEWAY
        serving_endpoint = os.environ.get("PREDIBASE_SERVING_ENDPOINT") or _DEFAULT_SERVING_ENDPOINT

        if not api_token:
            raise PredibaseError(
                "An api_token must either be explicitly provided or set via the PREDIBASE_API_TOKEN "
                "environment variable."
            )

        if not api_gateway:
            raise PredibaseError("PREDIBASE_GATEWAY cannot be empty.")

        if not api_gateway.startswith("https://"):
            logger.info(
                f"No HTTP scheme provided for PREDIBASE_GATEWAY (got: {api_gateway}), defaulting to `https://`."
            )
            api_gateway = f"https://{api_gateway}"

        if not serving_endpoint:
            raise PredibaseError("PREDIBASE_SERVING_ENDPOINT cannot be empty.")

        self.api_token = api_token
        self.api_gateway = api_gateway
        self.serving_http_endpoint = serving_endpoint

        self._session = start_session(self.api_gateway, self.api_token, self.serving_http_endpoint)

        self.adapters = Adapters(self)
        self.completions = Completions(self)
        self.datasets = Datasets(self._session)
        self.deployments = Deployments(self)
        self.models = Models(self._session)
        self.repos = Repos(self._session)

    @property
    def default_headers(self) -> Dict[str, str]:
        return {
            "Authorization": "Bearer " + self.api_token,
            "User-Agent": f"predibase-sdk/{__version__} ({platform.version()})",
        }

    def _get_current_user(self):
        resp = self.get_json("/users/current")
        return User.from_dict(resp)

    @warn_outdated_sdk
    def _get(self, endpoint: str, params: dict = None):
        return self._http.get(urljoin(self.api_gateway, endpoint), headers=self.default_headers, params=params)

    @warn_outdated_sdk
    def _post(self, endpoint: str, data: Optional[Any] = None, json: Optional[Any] = None, **kwargs):
        return self._http.post(
            urljoin(self.api_gateway, endpoint), data=data, json=json, headers=self.default_headers, **kwargs
        )

    @warn_outdated_sdk
    def _delete(self, endpoint: str):
        return self._http.delete(urljoin(self.api_gateway, endpoint), headers=self.default_headers)

    @warn_outdated_sdk
    def _put(self, endpoint: str, json: Optional[Any]):
        return self._http.put(urljoin(self.api_gateway, endpoint), json=json, headers=self.default_headers)

    @warn_outdated_sdk
    def _get_serving(self, endpoint: str):
        return self._http.get(urljoin(self.serving_http_endpoint, endpoint), headers=self.default_headers)

    @warn_outdated_sdk
    def _post_serving(self, endpoint: str, data: Optional[Any] = None, json: Optional[Any] = None, **kwargs):
        return self._http.post(
            urljoin(self.serving_http_endpoint, endpoint),
            data=data,
            json=json,
            headers=self.default_headers,
            **kwargs,
        )

    @warn_outdated_sdk
    def _get_ws(self, endpoint: str):
        ws_url = self.api_gateway.replace("https://", "wss://").replace("http://", "ws://")
        return requests.Session().get(ws_url + endpoint, headers=self.default_headers, timeout=None, stream=True)

    def _format_url(self, endpoint: str) -> str:
        if "localhost" in self.api_gateway:
            root_url = "http://localhost:8000"
        else:
            url = self.api_gateway
            if "api." in url:
                url = url.replace("api.", "").replace("/v1", "")
            root_url = url

        from urllib.parse import urljoin

        return urljoin(root_url, endpoint)



import concurrent.futures
import functools
import os
import time
from dataclasses import dataclass, field
from pprint import pprint
from string import Formatter
from typing import Any, Dict, List, Optional, TYPE_CHECKING, Union

import deprecation
from dataclasses_json import config, dataclass_json, LetterCase

from predibase.resource import model as mdl
from predibase.resource.connection import get_dataset
from predibase.resource.dataset import Dataset
from predibase.resource.engine import Engine
from predibase.resource.llm.prompt import PromptTemplate
from predibase.resource.llm.response import GeneratedResponse
from predibase.resource.llm.template import FinetuneTemplateCollection
from predibase.resource.llm.util import get_quantization_parameters, print_events
from predibase.resource_util import build_model
from predibase.util import load_yaml
from predibase.version import __version__

if TYPE_CHECKING:
    from predibase.pql.api import Session

_PATH_HERE = os.path.abspath(os.path.dirname(__file__))
_TEMPLATE_DIR = os.path.join(_PATH_HERE, "templates")


class HuggingFaceLLM:
    def __init__(self, session: "Session", model_name: str):
        self.session = session
        self.model_name = model_name

    def get_finetune_templates(self) -> FinetuneTemplateCollection:
        return FinetuneTemplateCollection(self)

    def deploy(
        self,
        deployment_name: str,
        engine_template: Optional[str] = None,
        hf_token: Optional[str] = None,
        auto_suspend_seconds: Optional[int] = None,
        max_input_length: Optional[int] = None,
        max_total_tokens: Optional[int] = None,
        max_batch_prefill_tokens: Optional[int] = None,
        revision: Optional[str] = None,
        quantization_kwargs: Optional[Dict[str, str]] = None,
        overwrite: Optional[bool] = False,
    ) -> "LLMDeploymentJob":
        return deploy_llm(
            session=self.session,
            deployment_name=deployment_name,
            engine_template=engine_template,
            hf_token=hf_token,
            auto_suspend_seconds=auto_suspend_seconds,
            max_input_length=max_input_length,
            max_total_tokens=max_total_tokens,
            max_batch_prefill_tokens=max_batch_prefill_tokens,
            quantization_kwargs=quantization_kwargs,
            revision=revision,
            huggingface_model_name=self.model_name,
            overwrite=overwrite,
        )

    def finetune(
        self,
        prompt_template: Optional[Union[str, PromptTemplate]] = None,
        target: Optional[str] = None,
        dataset: Optional[Union[str, Dataset]] = None,
        engine: Optional[Union[str, Engine]] = None,
        config: Optional[Union[str, Dict]] = None,
        repo: Optional[str] = None,
        epochs: Optional[int] = None,
        train_steps: Optional[int] = None,
        learning_rate: Optional[float] = None,
    ) -> "mdl.ModelFuture":
        if isinstance(prompt_template, PromptTemplate):
            prompt_template = prompt_template.render()

        if config is None:
            config = self.get_finetune_templates().default.to_config(prompt_template=prompt_template, target=target)
            # HACK(Arnav): Temporary hack to inject the right target_modules into the config for mixtral and
            # mixtral instruct. This should be removed once the Mixtral models have a default PEFT mapping for LoRA
            # target_modules. See MLX-1680: https://linear.app/predibase/issue/MLX-1680/remove-peftlora-mixtral-hack-in-the-predibase-sdk # noqa
            if (
                config.get("base_model", "") in {"mistralai/Mixtral-8x7B-v0.1", "mistralai/Mixtral-8x7B-Instruct-v0.1"}
                and "adapter" in config
                and config.get("adapter", {}).get("target_modules", None) is None
            ):
                config["adapter"]["target_modules"] = ["q_proj", "v_proj"]
        else:
            if isinstance(config, str):
                config = load_yaml(config)

            if not isinstance(config, dict):
                raise ValueError(f"Invalid config type: {type(config)}, expected str or dict")

        # Apply first-class training parameters.
        if train_steps is not None and epochs is not None:
            raise ValueError("Cannot specify both train_steps and epochs.")
        if train_steps is not None:
            config["trainer"]["train_steps"] = train_steps
        if epochs is not None:
            config["trainer"]["epochs"] = epochs
        if learning_rate is not None:
            config["trainer"]["learning_rate"] = learning_rate

        if repo is None:
            # If no repo is specified, automatically construct the repo name from the dataset and model name.
            dataset_name = dataset.name if isinstance(dataset, Dataset) else dataset
            if "/" in dataset_name:
                _, dataset_name = dataset_name.split("/")

            model_name = self.model_name
            if "/" in model_name:
                _, model_name = model_name.split("/")

            repo = f"{model_name}-{dataset_name}"

        if "/" in repo:
            repo = repo.replace("/", "-")

        repo: "mdl.ModelRepo" = get_or_create_repo(self.session, repo)
        if dataset is None:
            # Assume the dataset is the same as the repo head
            md = repo.head().to_draft()
            md.config = config
        else:
            if isinstance(dataset, str):
                conn_name = None
                if "/" in dataset:
                    conn_name, dataset = dataset.split("/")
                dataset = get_dataset(self.session, dataset, connection_name=conn_name)
            md = repo.create_draft(config=config, dataset=dataset)

        return md.train_async(engine=engine)


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class _LLMDeployment:
    id: int = field(metadata=config(field_name="id"))
    tenant_id: int = field(metadata=config(field_name="tenantID"))
    uuid: str = field(metadata=config(field_name="uuid"))
    name: str = field(metadata=config(field_name="name"))
    description: Optional[str] = field(metadata=config(field_name="description"))
    model_name: str = field(metadata=config(field_name="modelName"))
    adapter_name: Optional[str] = field(metadata=config(field_name="adapterName"))
    num_shards: Optional[int] = field(metadata=config(field_name="numShards"))
    quantize: str = field(metadata=config(field_name="quantize"))
    deployment_status: str = field(metadata=config(field_name="deploymentStatus"))

    prompt_template: str = field(metadata=config(field_name="promptTemplate"))
    min_replicas: int = field(metadata=config(field_name="minReplicas"))
    max_replicas: int = field(metadata=config(field_name="maxReplicas"))
    created: str = field(metadata=config(field_name="created"))
    updated: str = field(metadata=config(field_name="updated"))
    created_by_user_id: Optional[int] = field(metadata=config(field_name="createdByUserID"))
    scale_down_period: int = field(metadata=config(field_name="scaleDownPeriod"))
    is_shared: bool = field(metadata=config(field_name="isShared"), default=False)
    error_text: Optional[str] = field(metadata=config(field_name="errorText"), default=None)
    dynamic_adapter_loading_enabled: bool = field(
        metadata=config(field_name="dynamicAdapterLoadingEnabled"),
        default=False,
    )


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class LLMDeploymentReadyResponse:
    name: str
    ready: bool


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class LLMDeploymentScaledResponse:
    name: str
    scaled: bool


class LLMDeployment:
    def __init__(
        self,
        session: "Session",
        name: str,
        # adapter can be a HuggingFaceLLM or mdl.Model
        adapter: Optional[Union["mdl.Model", HuggingFaceLLM]] = None,
        deployment_metadata: Optional[_LLMDeployment] = None,
    ):
        self.session = session
        self.name = name

        # Before trying to deserialize the LLM, if the response is empty,
        # then we should raise an error telling the user the LLM does not exist.
        llm_json = self.session.get_json(f"/llms/{self.name}")
        if not llm_json:
            raise RuntimeError(f"Could not find LLM {name}")

        self.data = _LLMDeployment.from_dict(llm_json) if deployment_metadata is None else deployment_metadata

        if adapter:
            if not self.data.dynamic_adapter_loading_enabled:
                raise RuntimeError(f"Base deployment {self.name} is not configured to support LoRAX")

            # TODO: (magdy) check base model compatibility for hf based adapters here
            # (Delayed now since we need to do a roundtrip to hf) (INFRA-2052)
            # Directly import `Model` here to avoid cyclic import issues.
            from predibase.resource.model import Model

            if isinstance(adapter, Model) and not self._is_adapter_compatible(adapter):
                raise RuntimeError(
                    f"base deployment {self.name} does not match the adapter's base model. "
                    f"Expected base deployment model to be: {adapter.llm_base_model_name}. "
                    f"Actual: {self.data.model_name}",
                )

        self._adapter = adapter

    def _is_adapter_compatible(self, adapter: Optional[Union["mdl.Model", HuggingFaceLLM]]) -> bool:
        if self.data.model_name == adapter.llm_base_model_name:
            return True

        # It is possible that the base deployment's model name is not the same as the name of the base
        # model the adapter was trained on. This can happen if the adapter was trained on a model that
        # was subsequently renamed, or in the case where we're allowing users to LoRAX over dequantized
        # upscaled shared LLMs in Predibase.
        return self.session.get_json(
            "/llms/validate_adapter_compatibility",
            params={
                "deploymentBaseModelName": self.data.model_name,
                "adapterBaseModelName": adapter.llm_base_model_name,
            },
        )["compatible"]

    @property
    def adapter(self) -> Optional[Union["mdl.Model", HuggingFaceLLM]]:
        return self._adapter

    def with_adapter(self, model: Union["mdl.Model", HuggingFaceLLM]) -> "LLMDeployment":
        return LLMDeployment(
            session=self.session,
            name=self.name,
            adapter=model,
            deployment_metadata=self.data,
        )

    def generate(
        self,
        prompt: str,
        options: Optional[Dict[str, Union[str, float]]] = None,
    ) -> GeneratedResponse:
        data = {
            "inputs": prompt,
            "parameters": options if isinstance(options, dict) else dict(),
        }

        if self._adapter:
            if isinstance(self._adapter, HuggingFaceLLM):
                data["parameters"]["adapter_id"] = self._adapter.model_name
                data["parameters"]["adapter_source"] = "hub"
            else:
                data["parameters"][
                    "adapter_id"
                ] = f"{self._adapter.uuid}/{self._adapter.best_run_id}/artifacts/model/model_weights/"
                data["parameters"]["adapter_source"] = "s3"

        res = self.session.post_json_serving(
            endpoint=f"/{self.session.tenant}/deployments/v2/llms/{self.name}/generate",
            json=data,
        )
        return GeneratedResponse(
            prompt=prompt,
            response=res.get("generated_text", ""),
            model_name=self.name,
            generated_tokens=res.get("details", {}).get("generated_tokens", None),
            prefill=res.get("details", {}).get("prefill", None),
            tokens=res.get("details", {}).get("tokens", None),
            finish_reason=res.get("details", {}).get("finish_reason", None),
            seed=res.get("details", {}).get("seed", None),
            best_of_sequences=res.get("details", {}).get("best_of_sequences", None),
        )

    def generate_batch(
        self,
        prompts: List[str],
        options: Optional[Dict[str, Union[str, float]]] = None,
    ) -> List[GeneratedResponse]:
        if not isinstance(prompts, list):
            raise ValueError("`prompts` must be a List[str]")

        options = options if isinstance(options, dict) else dict()
        resp_list = []
        with concurrent.futures.ThreadPoolExecutor() as executor:
            futures = []
            for prompt in prompts:
                future = executor.submit(self.generate, prompt=prompt, options=options)
                futures.append(future)

            for future in concurrent.futures.as_completed(futures):
                try:
                    res = future.result()
                    resp_list.append(res)
                except Exception as exc:
                    print("ERROR:", exc)

        if len(resp_list) == 0:
            raise RuntimeError("LLM failed to generate a response. Please try again later.")
        return resp_list

    def prompt_batch(
        self,
        data_list: Union[List[str], List[Dict[str, Any]]],
        temperature: float = 0.1,
        max_new_tokens: int = 128,
        bypass_system_prompt: bool = False,
    ):
        if not isinstance(data_list, list):
            raise ValueError("`prompts` must be a List[str]")

        resp_list = []
        with concurrent.futures.ThreadPoolExecutor(max_workers=1000) as executor:
            futures = []
            for data in data_list:
                future = executor.submit(
                    self.prompt,
                    data=data,
                    temperature=temperature,
                    max_new_tokens=max_new_tokens,
                    bypass_system_prompt=bypass_system_prompt,
                )
                futures.append(future)

            for future in concurrent.futures.as_completed(futures):
                res = None
                try:
                    res = future.result()
                except Exception as exc:
                    print("ERROR:", exc)
                finally:
                    resp_list.append(res)

        if len(resp_list) == 0:
            raise RuntimeError("LLM failed to generate a response. Please try again later.")
        return resp_list

    def prompt(
        self,
        data: Union[str, Dict[str, Any]],
        temperature: float = 0.1,
        max_new_tokens: int = 128,
        bypass_system_prompt: bool = False,
    ) -> GeneratedResponse:
        template = self.default_prompt_template
        if template is None:
            if isinstance(data, str):
                template = "{__pbase_data_input__}"
            elif len(data) == 1:
                template = f"{{{next(iter(data))}}}"
            else:
                raise RuntimeError(
                    "Unable to interpolate multiple keys in data into prompt " "- no default_prompt_template exists",
                )

        fields_in_template = {tpl[1] for tpl in Formatter().parse(template) if tpl[1] is not None}

        if isinstance(data, str):
            if len(fields_in_template) > 1:
                raise RuntimeError(
                    "Only a single string was provided as prompt input, but there are multiple "
                    "interpolation fields in the `default_prompt_template`.",
                )

            # Coerce data into a 1-element dict matching the single interpolation field
            # in the prompt template.
            data = {next(iter(fields_in_template)): data}

        if len(fields_in_template.symmetric_difference(data.keys())) > 0:
            raise RuntimeError(
                f"Fields in prompt template do not match fields in data: "
                f"template: {fields_in_template}, data: {data.keys()}",
            )

        template = template.format(**data)

        deployment_ready = self.is_ready()
        if not deployment_ready:
            print(
                f"Target LLM deployment `{self.name}` is not active yet. This "
                f"call will block until the LLM is in a ready state.",
            )
            if self.data.is_shared:
                print("This might take a few seconds.")
            else:
                print(
                    "This might take a few minutes. Call `LLMDeployment.get_events()` "
                    "to get the status of your deployment.",
                )

            # block to wait for the LLM to spin up.
            self.wait_for_ready(timeout_seconds=1500)

        # Wrap with the system prompt template if one exists
        if self.data.prompt_template and not bypass_system_prompt:
            template = self.data.prompt_template % template

        # talk directly to the LLM, bypassing Temporal and the engines.
        responses = self.generate(
            template,
            options={"temperature": temperature, "max_new_tokens": max_new_tokens},
        )
        return responses

    @functools.cached_property
    def default_prompt_template(self) -> Optional[str]:
        if self._adapter:
            # Dynamic adapter deployment case
            from predibase.resource.model import Model

            if isinstance(self._adapter, Model):
                adapter: Model = self._adapter
                return adapter.config.get("prompt", {}).get("template", None)
            else:
                # Some arbitrary HuggingFace adapter is being used.
                return None
        elif self.data.adapter_name:
            # Dedicated fine-tuned deployment case
            # Adapter name is actually the path to the model weights and has the form:
            # "<model_uuid>/<model_best_run_id>/artifacts/model/model_weights/"
            model_uuid = self.data.adapter_name.split("/")[0]
            try:
                resp = self.session.get_json(f"/models/version/uuid/{model_uuid}")
                model = build_model(resp, self.session)
                return model.config.get("prompt", {}).get("template", None)
            except Exception as e:
                raise RuntimeError("Failed to get info on registered adapter") from e
        else:
            # Base OSS LLM deployment case
            return None

    def delete(self):
        print(f"Requested deletion of llm deployment: `{self.name}` ...")
        endpoint = f"/llms/{self.name}"
        self.session.delete_json(endpoint)

        while True:
            try:
                llm_deployment = self.session.get_json(endpoint)
                if llm_deployment is None:
                    print(f"Successfully deleted llm deployment: `{self.name}`")
                    break
            except Exception as e:
                raise RuntimeError(f"Error while deleting deployment `{self.name}`: {e} {type(e)}.")
            time.sleep(1.0)

    def is_ready(self) -> bool:
        try:
            self.session.get_json_serving(f"/{self.session.tenant}/deployments/v2/llms/{self.name}/health")
            return True
        except Exception:
            return False

    def wait_for_ready(self, timeout_seconds: int = 600, poll_interval_seconds: int = 5) -> bool:
        start = time.time()
        while int(time.time() - start) < timeout_seconds:
            if self.is_ready():
                return True
            time.sleep(poll_interval_seconds)
        return False

    def is_scaled(self) -> bool:
        resp = self.session.get_json(f"/llms/{self.name}/scaled")
        return LLMDeploymentScaledResponse.from_dict(resp).scaled

    @deprecation.deprecated(
        deprecated_in="2023.12.8",
        current_version=__version__,
        details="Use the LLMDeployment().get_events method instead to get information about the deployment.",
    )
    def get_status(self) -> str:
        resp = _LLMDeployment.from_dict(self.session.get_json(f"/llms/{self.name}"))
        return resp.deployment_status

    def get_events(self, detailed: Optional[bool] = False):
        detailed = str(detailed).lower()
        events = self.session.get_json(f"/llms/{self.name}/events?detailed={detailed}")
        if detailed == "false":
            print_events(events, print_header=True)
        return events


class LLMDeploymentJob:
    def __init__(self, deployment_name: str, session: "Session"):
        self._deployment_name = deployment_name
        self._uri = f"pb://jobs/deploy::{deployment_name}"
        self._session = session

    def get(self) -> LLMDeployment:
        self._session.get_llm_deployment_events_until_with_logging(
            events_endpoint=f"/llms/{self._deployment_name}/events?detailed=false",
            success_cond=lambda resp: "Ready" in [r.get("eventType", None) for r in resp.get("ComputeEvents", [])],
            error_cond=lambda resp: "BackOff" in [r.get("eventType", None) for r in resp.get("ComputeEvents", [])]
            or resp.get("deploymentStatus", None) in ("failed", "canceled", "stopped"),
        )
        resp = self._session.get_json(f"/llms/{self._deployment_name}")
        return LLMDeployment(self._session, self._deployment_name, deployment_metadata=_LLMDeployment.from_dict(resp))

    def cancel(self):
        return self._session.post_json(f"/llms/{self._deployment_name}/cancel", {})


def get_or_create_repo(session: "Session", repo_name: str) -> "mdl.ModelRepo":
    return mdl.create_model_repo(session, name=repo_name, exists_ok=True)


def deploy_llm(
    session: "Session",
    deployment_name: str,
    engine_template: Optional[str] = None,
    hf_token: Optional[str] = None,
    auto_suspend_seconds: Optional[int] = None,
    max_input_length: Optional[int] = None,
    max_total_tokens: Optional[int] = None,
    max_batch_prefill_tokens: Optional[int] = None,
    quantization_kwargs: Optional[Dict[str, str]] = None,
    huggingface_model_name: Optional[str] = None,
    revision: Optional[str] = None,
    predibase_model_uri: Optional[str] = None,
    overwrite: Optional[bool] = False,
) -> "LLMDeploymentJob":
    """Deploy a base or fine-tuned LLM using specified parameters.

    Args:
        session (Session): The session object for interacting with the server.
        deployment_name (str): The name of the deployment in Predibase (e.g. 'my-llama-deployment').
        engine_template (Optional[str]): The engine template to use for deployment.
        hf_token (Optional[str]): Huggingface API token.
        auto_suspend_seconds (Optional[int]): Number of seconds before autoscaling down (0 for no autoscaling).
        max_input_length (Optional[int]): Maximum input length for the LLM.
        max_total_tokens (Optional[int]): This is the token budget that the model will have per request.
            As an example, if `max_total_tokens` is 1000 tokens and the prompt length (in tokens) is 800,
            the model will only be able to generate new 200 tokens. Since this is a per-request parameter,
            the larger this value, the larger amount each request will take in the GPU, and the less effective
            batching can be as larger values will allow for fewer batches.
        max_batch_prefill_tokens (Optional[int]): An upper bound on the number of tokens for the prefill operation.
        quantization_kwargs (Optional[Dict[str, str]]): Quantization parameters for model.
            See `get_quantization_parameters` for more details on the format.
        revision (Optional[str]): Huggingface model revision.
        huggingface_model_name (Optional[str]): Huggingface model name (e.g. 'meta-llama/Llama-2-7b-hf').
        predibase_model_uri (Optional[str]): Predibase model URI for deployment (e.g. pb://models/my_repo/2).

    Returns:
        LLMDeploymentJob: An object representing an LLM deployment job.

    Raises:
        ValueError:
            - If exactly one of 'huggingface_model_name' and 'predibase_model_uri' is not provided.
            - If 'quantization_kwargs' are not valid per 'get_quantization_parameters'.
    """
    # TODO: if model is an adapter add the base model from the adapter_config.json
    # Validate that exactly one of them is non-None
    if bool(huggingface_model_name) != bool(predibase_model_uri):
        if huggingface_model_name is not None:
            print(f"Deploying Huggingface model: {huggingface_model_name}")
        else:
            print(f"Deploying Predibase model: {predibase_model_uri}")
    else:
        raise ValueError("Exactly one of 'huggingface_model_name' and 'predibase_model_uri' must be provided.")

    # Extract quantization parameters from user-provided dictionary.
    quantize, environment_variables = get_quantization_parameters(quantization_kwargs)

    # Parse the generation parameters into customArgs.
    custom_args = [
        [arg[0], str(arg[1])]
        for arg in [
            ("--max-input-length", max_input_length or 1024),
            ("--max-total-tokens", max_total_tokens or 2048),
            ("--max-batch-prefill-tokens", max_batch_prefill_tokens or 4096),
        ]
        if arg[1] is not None
    ]
    custom_args = [element for sublist in custom_args for element in sublist]

    # Put together the deployment parameters.
    auto_suspend_seconds = 3600 if auto_suspend_seconds is None else auto_suspend_seconds

    model_params = {
        "name": deployment_name,
        "modelName": huggingface_model_name,
        "modelUri": predibase_model_uri,
        "engineTemplate": engine_template,
        "scaleDownPeriod": auto_suspend_seconds,
        "source": "huggingface",
        "customArgs": custom_args,
        "quantize": quantize,
        "environmentVariables": environment_variables,
        "hfToken": hf_token,
        "modelRevision": revision,
    }
    # filter None params as those can't be decoded by the server.
    model_params = {key: value for key, value in model_params.items() if value is not None}

    if overwrite:
        # Make the put request to update the deployment.
        llm_deployment_params = session.put_json(f"/llms/{deployment_name}", json=model_params)
    else:
        # Make the post request to create the deployment.
        llm_deployment_params = session.post_json("/llms", json=model_params)

    # Print information to the user.
    base_msg = "Important: This is a dedicated LLM deployment for which Predibase charges per hour of usage."
    if auto_suspend_seconds == 0:
        msg = "It will remain active until you choose to delete the deployment."
    else:
        msg = (
            f"It will remain active until the autoscaling scales it down after {auto_suspend_seconds} "
            f"seconds of inactivity. You can also delete the deployment to completely shut it down."
        )
    print(base_msg + " " + msg + "\n")
    print("Deploying the model with the following params:")
    pprint(llm_deployment_params)

    # Return LLMDeploymentJob.
    return LLMDeploymentJob(deployment_name, session)



from __future__ import annotations

from predibase.pql import Session
from predibase.resource.model import create_model_repo, ModelRepo


class Repos:
    def __init__(self, session: Session):
        self._session = session

    def create(self, name: str, description: str | None = None, exists_ok: bool = False) -> ModelRepo:
        return create_model_repo(
            self._session,
            name,
            description,
            engine=None,
            exists_ok=exists_ok,
        )



from typing import Dict, List, Optional, Union

import pandas as pd

from predibase.pql.api import ServerResponseError
from predibase.resource.engine import DefaultEngineType, Engine
from predibase.util import log_info

SUPPORTED_SERVICE_TYPES = ["general", "serving"]


class EngineMixin:
    def create_engine(
        self,
        name,
        template: Optional[str] = None,
        template_id: Optional[int] = None,  # deprecated
        environment_id: Optional[int] = None,
        preload_model_path: Optional[str] = None,
        auto_suspend: int = 600,
        auto_resume: bool = True,
        exists_ok: bool = False,
    ) -> Engine:
        """
        :param auto_resume: whether the engine is allowed to automatically resume from a suspended state
            when given a job
        :param auto_suspend: how long the engine can be idle (have no active jobs) before resources are scaled down.
            Suspended engines take 8-10 minutes to start up again
        :param name: name of engine
        :param template: template name to use when creating the engine (see: get_engine_templates())
        :param template_id: template_id obtained from get_engine_templates() (DEPRECATED)
        :param environment_id: environment_id obtained from get_environments()
        :param preload_model_path: path to model to preload into the engine
        :param exists_ok: Do not raise an error and return the existing engine.
        :return: Engine
        """
        template_id = self._get_template_id(template) if template_id is None else template_id

        # TODO(geoffrey): should we make the user specify the service_type instead of inferring it here?
        service_type = self.get_engine_templates(full_info=True)[template_id]["serviceType"]
        if service_type == "serving":
            log_info("Creating engine of type 'serving'. Ignoring auto_suspend and auto_resume settings.")

        body = {
            "name": name,
            "templateID": template_id,
            "serviceType": service_type,
            "autoSuspendSeconds": auto_suspend,
            "autoResume": auto_resume,
        }
        if environment_id is not None:
            body["environmentID"] = environment_id
        if preload_model_path is not None:
            body["preloadModelPath"] = preload_model_path
        try:
            resp = self.session.post_json("/engines", body)
            return Engine.from_dict({"session": self.session, **resp})
        except ServerResponseError as e:
            if exists_ok and e.code == 400:
                log_info(f"Engine {name} already exists. exists_ok=True, so ignoring.")
                return self.get_engine(name)
            else:
                raise e

    def get_engine(
        self,
        name: Optional[str] = None,
        engine_id: Optional[int] = None,
        activate: Optional[bool] = False,
    ) -> Engine:
        if not (engine_id or name):
            raise ValueError("Must provide either engine_name or engine_id")
        if name and engine_id:
            raise ValueError("Cannot provide both engine name and engine_id")

        endpoint = f"/engines/{engine_id}" if engine_id else f"/engines/name/{name}"
        resp = self.session.get_json(endpoint)
        engine = Engine.from_dict({"session": self.session, **resp})
        if activate:
            try:
                engine.start()
            except Exception as e:
                print(f"Failed to start engine {name}: {e}")
        return engine

    def set_current_engine(
        self,
        name: Optional[str] = None,
        engine_id: Optional[int] = None,
        engine_type: DefaultEngineType = DefaultEngineType.QUERY,
    ):
        if engine_id and name:
            raise ValueError("Cannot provide both engine name and engine_id")
        if name:
            endpoint = f"/engines/current/name/{name}/{engine_type.value}"
            resp = self.session.put_json(endpoint, {})
        else:
            endpoint = f"/engines/current/{engine_type.value}"
            resp = self.session.put_json(endpoint, {"id": engine_id})
        return resp

    def get_current_engine(self, engine_type: DefaultEngineType = DefaultEngineType.QUERY):
        endpoint = "/engines/current"
        resp = self.session.get_json(endpoint)
        if engine_type.value not in resp:
            raise ValueError(f"Engine type {engine_type.value} not found in response: {resp.keys()}")
        return Engine.from_dict({"session": self.session, **resp[engine_type.value]})

    def list_engines(self, df: bool = False) -> Union[pd.DataFrame, List[Engine]]:
        """
        :param df: Whether to return pandas dataframe or list of objects
        :return: Pandas dataframe or list of engines
        """
        resp = self.session.get_json("/engines")
        engines = [x for x in resp["engines"]]
        if df:
            return pd.DataFrame(engines)
        return [Engine.from_dict({"session": self.session, **x}) for x in engines]

    def delete_engine(self, name: str):
        endpoint = f"/engines/name/{name}"
        resp = self.session.delete_json(endpoint)
        return resp

    def get_engine_templates(self, full_info: bool = False) -> Union[Dict[int, Dict[str, str]], Dict[int, Dict]]:
        """
        :param full_info: whether to return full info (with node labels, cpu, memory, storage, etc)
        :return: Dict of template_id -> template spec (i.e. "cpu-xsmall")
        """
        resp = self.session.get_json("/engines/schema")
        templates = resp["templates"]
        if full_info:
            return {x["id"]: x for x in templates}
        return {x["id"]: {"name": x["name"], "description": x["description"]} for x in templates}

    def get_engine_environments(self) -> Dict[int, Dict]:
        """
        :return: Dict of env UUID -> env spec (i.e. "cpu-xsmall")
        """
        resp = self.session.get_json("/environments/available")
        environments = resp["environments"]
        return {x["id"]: {"name": x["name"], "type": x["type"], "id": x["id"], "uuid": x["uuid"]} for x in environments}

    def _template_name_to_ids(self) -> Dict[str, int]:
        id_to_meta = self.get_engine_templates()
        return {meta["name"]: _id for _id, meta in id_to_meta.items()}

    def _get_template_id(self, name: str) -> int:
        if name is None:
            return 1
        name_to_ids = self._template_name_to_ids()
        if name not in name_to_ids:
            raise ValueError(
                f"Engine template name '{name}' not found. Available templates: {list(name_to_ids.keys())}",
            )
        return name_to_ids[name]



from typing import Optional

import typer

from predibase.cli_commands.utils import get_client, get_console

app = typer.Typer(no_args_is_help=True)


@app.command()
def dataset(
    file_path: str = typer.Option(
        None,
        "--file-path",
        "-f",
        prompt="Path of file to upload",
        prompt_required=True,
        help="Path of the dataset file to be uploaded",
    ),
    name: Optional[str] = typer.Option(
        None,
        "--name",
        "-n",
        prompt="Dataset name",
        prompt_required=False,
        help="Name of the dataset to be created - will be auto-generated if left unset",
    ),
):
    client = get_client()

    get_console().print(f"Uploading file at {file_path} as a dataset...")

    ds = client.upload_dataset(file_path, name)

    get_console().print(f"Created dataset {ds.name}")


if __name__ == "__main__":
    app()



import copy
import datetime
import os
import tempfile
import time
import zipfile
from collections import defaultdict
from dataclasses import dataclass, field
from functools import cached_property
from threading import Event, Thread
from typing import Any, Dict, List, Optional, Tuple, Union
from urllib.parse import urlparse

import pandas as pd
import requests
from dataclasses_json import config, dataclass_json, LetterCase
from dateutil import parser

from predibase.pql.api import ServerResponseError, Session
from predibase.pql.utils import get_results_df
from predibase.resource.config import ConfigSuggestion
from predibase.resource.dataset import DataField, Dataset

# if TYPE_CHECKING:  # Only imports the below statements during type checking
#     from predibase.dataset import Dataset
from predibase.resource.engine import Engine
from predibase.resource.llm import interface
from predibase.resource.user import User
from predibase.resource.viz import VisualizeType
from predibase.resource_util import build_model_repo
from predibase.util import get_url, log_info, spinner
from predibase.util.metrics import is_llm_model
from predibase.util.tensorboard import launch_tensorboard as launch_tb
from predibase.util.url import encode_url_param

TRAIN_METRICS = "train_metrics"
VALIDATION_METRICS = "validation_metrics"
TEST_METRICS = "test_metrics"

MODEL_WEIGHTS = "model_weights"


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class ModelRepo:
    session: Session
    _id: str = field(metadata=config(field_name="id"))
    _uuid: str = field(metadata=config(field_name="uuid"))
    _name: str = field(metadata=config(field_name="modelName"))
    _created: str = field(metadata=config(field_name="created"))
    _updated: str = field(metadata=config(field_name="updated"))
    _description: Optional[str] = field(metadata=config(field_name="description"), default=None)
    _root_id: Optional[str] = field(metadata=config(field_name="rootID"), default=None)
    _user: Optional[User] = field(metadata=config(field_name="user"), default=None)
    _parent_id: Optional[str] = field(metadata=config(field_name="parentID"), default=None)
    # Latest dataset
    _latest_config: Optional[dict] = field(metadata=config(field_name="latestConfig"), default=None)
    _latest_dataset_id: Optional[int] = field(metadata=config(field_name="datasetID"), default=None)
    _latest_dataset: Optional[Dataset] = field(metadata=config(field_name="dataset"), default=None)

    def __repr__(self):
        return (
            f"ModelRepo(id={self.id}, name={self._name}, description={self.description}, "
            f"latest_config={{...}}, "
            f"latest_dataset={self.latest_dataset}, "
            f"created={self._created}, updated={self._updated})"
        )

    def _build_model_with_backref(self, resp):
        from predibase.resource_util import build_model

        resp["repo"] = self
        return build_model(resp, self.session)

    @property
    def id(self):
        return self._id

    @property
    def uuid(self):
        return self._uuid

    @property
    def name(self):
        return self._name

    @property
    def description(self):
        return self._description

    @property
    def latest_config(self):
        return self._latest_config

    @property
    def latest_dataset(self):
        if self._latest_dataset:
            return self._latest_dataset
        # get_model().repo.latest_dataset isn't seeded
        if self._latest_dataset_id:
            resp = self.session.get_json(f"/datasets/{self._latest_dataset_id}")
            from predibase.resource_util import build_dataset

            self._latest_dataset = build_dataset(resp, self.session)
            return self._latest_dataset
        return None

    @property
    def latest_dataset_id(self):
        return self._latest_dataset_id

    @property
    def created(self):
        return self._created

    @property
    def updated(self):
        return self._updated

    @property
    def parent_id(self):
        return self._parent_id

    @property
    def root_id(self):
        return self._root_id

    @property
    def author(self):
        return self._user.username if self._user else None

    def create_draft(
        self,
        dataset: Optional[Dataset] = None,
        target: Optional[Union[str, List[str]]] = None,
        automl: Optional[bool] = False,
        hyperopt: Optional[bool] = False,
        config: Optional[dict] = None,
        description: Optional[str] = None,
    ) -> "ModelDraft":
        dataset = dataset or self.latest_dataset
        if dataset is None:
            raise ValueError(
                "Model repo has no latest dataset, and no dataset was provided. "
                "Cannot create model draft without a dataset.",
            )

        # need target to get the default config.
        if not config and target:
            config = get_default_config(self.session, dataset, target, automl, hyperopt)
        return ModelDraft(
            session=self.session,
            repo_id=self.id,
            _repo=self,
            name=self._name,
            config=config,
            dataset=dataset,
            description=description,
        )

    def _filter_model_weights(self, filename: str) -> bool:
        return "readme" not in filename.lower()

    def upload_model(self, model_path: str, dataset: Optional[Dataset] = None) -> "Model":
        required_metadata_files = ["model_hyperparameters.json", "training_progress.json", "training_set_metadata.json"]

        dataset_id = dataset.id if dataset is not None else self.latest_dataset_id
        if dataset_id is None:
            raise ValueError("must either provide a dataset or upload to a model repo with a dataset configured")

        with tempfile.TemporaryDirectory() as tmpdir:
            # If the path specified is a local folder, no need to unzip.
            model_artifacts_path = model_path
            if not os.path.isdir(model_path):
                model_artifacts_path = tmpdir
                with zipfile.ZipFile(model_path, "r") as zf:
                    zf.extractall(path=tmpdir)

            # Some model zips may save the files of interest in a model/ sub-directory.
            if os.path.exists(os.path.join(model_artifacts_path, "model")):
                model_artifacts_path = os.path.join(model_artifacts_path, "model")

            for required_file in required_metadata_files + [MODEL_WEIGHTS]:
                if not os.path.exists(os.path.join(model_artifacts_path, required_file)):
                    raise FileNotFoundError(
                        f"required file {required_file} not found in model folder {model_artifacts_path}",
                    )

            if not os.path.exists(os.path.join(model_artifacts_path, MODEL_WEIGHTS)):
                raise FileNotFoundError(f"model_weights file(s) not found in model folder {model_artifacts_path}")

            model_weights_files = []
            if os.path.isdir(os.path.join(model_artifacts_path, MODEL_WEIGHTS)):
                for file in os.listdir(os.path.join(model_artifacts_path, MODEL_WEIGHTS)):
                    if self._filter_model_weights(file):
                        model_weights_files.append(file)
            else:
                model_weights_files = [MODEL_WEIGHTS]

            metadata_zip_path = os.path.join(tmpdir, "model_metadata.zip")
            with zipfile.ZipFile(metadata_zip_path, "w") as metadata_zip:
                for f in required_metadata_files:
                    metadata_zip.write(os.path.join(model_artifacts_path, f), arcname=f)

            with open(metadata_zip_path, "rb") as zipped_metadata:
                files = {"file": zipped_metadata}
                file_name, _ = os.path.splitext(os.path.basename(metadata_zip_path))
                s = ","
                data = {
                    "fileName": file_name,
                    "datasetID": dataset_id,
                    "repoID": self.id,
                    "modelWeightsFiles": s.join(model_weights_files),
                }
                register_model_resp = self.session.post("/models/register_uploaded_model", files=files, data=data)

            model_dict = register_model_resp["model"]
            presigned_urls = register_model_resp["modelWeightsPresignedFiles"]

            # Loop through the model weight files and upload them to blob storage.
            # Beacuse we send the list of files to upload to the backend, we should
            # be able to rely on the order of the response matching our list.
            for idx, model_weight_file in enumerate(model_weights_files):
                presigned_url = presigned_urls[idx]
                upload_url = presigned_url["url"]
                upload_required_headers = presigned_url.get("headers", {})
                if upload_required_headers is None:
                    upload_required_headers = {}

                content_type = (
                    "application/json" if model_weight_file.lower().endswith(".json") else "binary/octet-stream"
                )
                headers = {"Content-Type": content_type}
                for k, v in upload_required_headers:
                    headers[k] = v

                file_path = (
                    os.path.join(model_artifacts_path, model_weight_file)
                    if model_weight_file == MODEL_WEIGHTS
                    else os.path.join(model_artifacts_path, "model_weights", model_weight_file)
                )
                with open(file_path, "rb") as weights_file:
                    # Upload file to blob storage with pre-signed url
                    requests.put(upload_url, data=weights_file, headers=headers).raise_for_status()

            # Signal successful upload of model weights.
            self.session.post(f"/models/complete_upload_model/{model_dict['id']}")

            model_dict["status"] = "ready"
            return self._build_model_with_backref(model_dict)

    def list_models(self, df: bool = False):
        endpoint = f"/models/repo/{self._id}?withVersions=true"
        resp = self.session.get_json(endpoint)
        if df:
            return pd.DataFrame([m for m in resp["modelRepo"]["models"]])
        return [self._build_model_with_backref(m) for m in resp["modelRepo"]["models"]]

    def draft_from_latest_version(self) -> "ModelDraft":
        return ModelDraft(
            session=self.session,
            repo_id=self.id,
            _repo=self,
            name=self._name,
            config=copy.deepcopy(self.latest_config),
            dataset=self.latest_dataset,
        )

    def latest(self) -> "Model":
        return get_model(self.session, self.name)

    def head(self) -> "Model":
        return self.latest()

    def get(self, version: Optional[Union[int, str]] = None) -> "Model":
        return get_model(self.session, self.name, version)


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class ModelRunMetadata:
    _id: str = field(metadata=config(field_name="id"))
    _model_id: int = field(metadata=config(field_name="modelID"))
    _run_id: str = field(metadata=config(field_name="runID"))
    _rendered_config: dict = field(metadata=config(field_name="renderedConfig"))

    @property
    def rendered_config(self) -> dict:
        return self._rendered_config


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class Model:
    # Disable the dataclass-json violation (ref: https://predibase.slack.com/archives/C02AZ1USR1U/p1696446768000669)
    # pylint: disable=used-before-assignment
    session: Session
    _id: str = field(metadata=config(field_name="id"))
    _uuid: str = field(metadata=config(field_name="uuid"))
    _repo_id: int = field(metadata=config(field_name="repoID"))
    _repo: ModelRepo = field(metadata=config(field_name="repo"))
    _config: dict = field(metadata=config(field_name="config"))
    _dataset_id: int = field(metadata=config(field_name="datasetID"))
    _dataset: Dataset = field(metadata=config(field_name="dataset"))
    _version: str = field(metadata=config(field_name="repoVersion"))
    _status: str = field(metadata=config(field_name="status"))
    _starred: bool = field(metadata=config(field_name="starred"))
    _archived: bool = field(metadata=config(field_name="archived"))
    # _created: datetime.datetime = config(metadata=config(field_name="created"))
    _created: str = field(metadata=config(field_name="created"))
    _user: Optional[User] = field(metadata=config(field_name="user"), default=None)
    _created_by_user_id: Optional[int] = field(metadata=config(field_name="createdByUserID"), default=None)
    _engine_id: Optional[str] = field(metadata=config(field_name="engineID"), default=None)
    _engine: Optional[Engine] = field(metadata=config(field_name="engine"), default=None)
    _engine_template: Optional[Dict[str, Any]] = field(metadata=config(field_name="engineTemplate"), default=None)
    _description: Optional[str] = field(metadata=config(field_name="description"), default=None)
    _experiment_id: Optional[int] = field(metadata=config(field_name="experimentID"), default=None)
    _error_text: Optional[str] = field(metadata=config(field_name="errorText"), default=None)
    _best_run_id: Optional[str] = field(metadata=config(field_name="bestRunID"), default=None)
    _active_run_id: Optional[str] = field(metadata=config(field_name="activeRunID"), default=None)
    _completed: Optional[str] = field(metadata=config(field_name="completed"), default=None)
    _training_created: Optional[str] = field(metadata=config(field_name="trainingCreated"), default=None)
    _training_completed: Optional[str] = field(metadata=config(field_name="trainingCompleted"), default=None)
    _active_run_meta: Optional[ModelRunMetadata] = field(metadata=config(field_name="modelRun"), default=None)
    _llm_base_model_name: Optional[str] = field(metadata=config(field_name="llmBaseModelName"), default=None)

    def __repr__(self):
        dataset = f"Dataset({self._dataset.name}...)"
        repo = f"Repo({self.repo.name}...)"
        return (
            f"Model(id={self._id}, repo={repo}, description={self.description}, "
            f"dataset={dataset}, engine=Engine({self.engine.name if self.engine else None}...), "
            f"config={{...}}, version={self.version}, "
            f"status={self.status}, created={self.created}, completed={self.completed})"
        )

    @property
    def id(self):
        return self._id

    @property
    def uuid(self) -> str:
        return self._uuid

    @property
    def best_run_id(self) -> str:
        return self._best_run_id

    @property
    def description(self):
        return self._description

    @property
    def config(self):
        return self._config

    @property
    def rendered_config(self):
        if self._active_run_meta is None:
            return None
        return self._active_run_meta.rendered_config

    @property
    def version(self):
        return self._version

    @property
    def repo_id(self):
        return self._repo_id

    @property
    def author(self):
        return self._user.username if self._user else None

    @property
    def repo(self):
        return self._repo

    @property
    def dataset_id(self):
        return self._dataset_id

    @property
    def engine_id(self):
        return self._engine_id

    @property
    def engine(self):
        if self._engine:
            return self._engine
        resp = self.session.get_json(f"/engines/{self.engine_id}")
        engine = Engine.from_dict({"session": self.session, **resp})
        self._engine = engine
        return engine

    @property
    def engine_template(self):
        return self._engine_template

    @property
    def dataset(self):
        return self._dataset

    @property
    def status(self):
        return self._get()["status"]

    @property
    def error_text(self) -> Union[str, None]:
        resp = self._get()
        if "errorText" in resp:
            return resp["errorText"]
        return None

    @property
    def active_fields(self) -> List[DataField]:
        resp = self._get()
        return [DataField.from_dict({"session": self.session, **x["field"]}) for x in resp["activeFields"]]

    @property
    def starred(self) -> bool:
        return self._starred

    @property
    def archived(self) -> bool:
        return self._archived

    @property
    def created(self) -> datetime.datetime:
        return pd.to_datetime(self._created)

    @property
    def completed(self) -> datetime.datetime:
        if self._completed:
            return pd.to_datetime(self._completed)
        resp = self._get()
        if "completed" in resp:
            self._completed = resp["completed"]
            return pd.to_datetime(self._completed)
        return None

    @property
    def training_created(self) -> datetime.datetime:
        if self._training_created:
            return self._training_created
        resp = self._get()
        if "trainingCreated" in resp:
            self._training_created = resp["trainingCreated"]
            return self._training_created
        return None

    @property
    def training_completed(self) -> datetime.datetime:
        if self._training_completed:
            return self._training_completed
        resp = self._get()
        if "trainingCompleted" in resp:
            self._training_completed = resp["trainingCompleted"]
            return self._training_completed
        return None

    @cached_property
    def metrics(self) -> "ModelMetrics":
        nested_metrics = {
            TRAIN_METRICS: defaultdict(dict),
            VALIDATION_METRICS: defaultdict(dict),
            TEST_METRICS: defaultdict(dict),
        }

        global_metrics = {
            TRAIN_METRICS: defaultdict(dict),
            VALIDATION_METRICS: defaultdict(dict),
            TEST_METRICS: defaultdict(dict),
        }

        general_metrics = {}

        for metric in self.active_run.data.metrics:
            parts = metric.key.split(".")

            if len(parts) == 1:
                general_metrics[metric.key] = metric.value
                continue

            # we'd like to capture the best metrics of the run.
            if parts[0] != "best" or len(parts) < 2:
                continue

            # if this field contains metrics, its format will look like the following
            # "best.train_metrics.Survived.accuracy"
            _, split, feature, name = parts

            if feature != "combined":
                nested_metrics[split][feature][name] = metric.value
            else:
                global_metrics[split][feature][name] = metric.value

        return ModelMetrics(nested_metrics, global_metrics, general_metrics)

    @property
    def active_run(self):
        return self.runs[self._active_run_id]

    @cached_property
    def runs(self) -> Dict[str, "ModelRun"]:
        resp = self.session.get_json(f"/models/version/{self.id}?withRuns=true")
        runs: List[ModelRun] = [ModelRun.from_dict(run) for run in resp["runs"]]
        return {run.info.run_id: run for run in runs}

    @property
    def llm_base_model_name(self) -> Optional[str]:
        return self._llm_base_model_name

    def _get(self):
        endpoint = f"/models/version/{self.id}"
        resp = self.session.get_json(endpoint)
        return resp["modelVersion"]

    def update_description(self, description: str):
        endpoint = f"/models/version/{self._id}"
        self.session.put_json(
            endpoint,
            {
                "id": self.id,
                "description": description,
                "starred": self.starred,
                "archived": self.archived,
            },
        )

    def delete(self):
        return self.session.delete_json(f"/models/version/{self.id}")

    def cancel(self):
        return self.session.post_json(f"/models/version/{self.id}/cancel", {})

    def _model_download_endpoint(self, framework: str) -> str:
        endpoint = f"/models/download/{self.id}"
        return endpoint + f"?framework={framework.lower()}"

    @spinner(name="exporting model")
    def _poll_model_download(self, endpoint: str) -> Dict[str, Any]:
        while True:
            resp = self.session.get_json(endpoint)
            status = resp["status"]
            if status == "exporting":
                time.sleep(2)
            elif status == "complete":
                break
            else:
                raise ValueError(f"Error exporting model: {resp['failure_reason']}")

        return resp

    def _start_model_download(self, endpoint: str) -> Dict[str, Any]:
        try:
            self.session.post(endpoint, {})
        except ServerResponseError as e:
            raise e

        return self._poll_model_download(endpoint)

    def download(
        self,
        name: Optional[str] = None,
        location: Optional[str] = None,
        framework: Optional[str] = None,
    ):
        """
        This method downloads the model to the specified location on your local machine.
        Args:
            name: Optional[str] - Name for the downloaded file.
            location: Optional[str] - File path to download the model to.
            framework: Optional[str] - Framework to download the model as. Default is to download as a Ludwig Model.

        Returns:
            None
        """
        if framework is None:
            framework = "ludwig"

        if is_llm_model(self._config) and framework != "ludwig":
            raise ValueError("LLMs can only be exported using the Ludwig format.")

        if location is not None:
            if not os.path.exists(location):
                raise ValueError(f"Location {location} does not exist.")

        endpoint = self._model_download_endpoint(framework)

        log_info("Downloading model")
        try:
            resp = self.session.get_json(endpoint)
        except ServerResponseError as e:
            # Model has not been previously exported. Start process.
            if "error 400: record not found" in e.message.lower():
                log_info("Model has not been exported yet. Starting export.")
                resp = self._start_model_download(endpoint)
                pass
            else:
                raise e

        status = resp["status"]
        # Wait for model to finish exporting
        if status == "exporting":
            resp = self._poll_model_download(endpoint)
        # If model was previously exported but failed, restart process
        elif status == "failed":
            resp = self._start_model_download(endpoint)

        # Now that model is successfully exported, download it
        url = resp["url"]
        download_path = None
        if name:
            if name[-4:] == ".zip":
                download_path = name
            else:
                download_path = name + ".zip"
        else:
            download_path = os.path.basename(urlparse(url).path)

        if location is not None:
            download_path = os.path.join(location, download_path)

        resp = requests.get(url)
        with open(download_path, "wb") as f:
            f.write(resp.content)

        log_info(f"Model downloaded to {download_path}")

    def export(
        self,
        location: str,
        connection: Optional[Union[str, int]] = None,
        framework: Optional[str] = None,
    ):
        """Export a model to an export location using the connection or target connection with the optional
        framework.

        :param location: Location as string.
        :param connection: (Optional[str, int]): Optional connection id or name for export location.
        :param framework: Optional framework, defaults to "ludwig"
        """
        connection_id = self.session.get_connection_id(connection) if connection else None
        query = f"EXPORT MODEL {self.repo.name} VERSION {self.version}"
        if framework is not None:
            query = query + f" WITH FORMAT (framework='{framework}')"
        query = query + f" TO '{location}'"
        return self.session.execute(query, connection_id=connection_id)

    def evaluate(
        self,
        targets: Union[str, List[str]],
        source: Union[Dataset, pd.DataFrame, str],
        limit: Optional[int] = None,
        metadata: Optional[bool] = False,
    ) -> pd.DataFrame:
        """Evaluate a model given a source dataset for one or more targets.

        :param targets (Union[str, List[str], Dict[str, str]]): One or more target names to make predictions on.
        :param source (Union[Dataset, pd.DataFrame, str]): Source dataset to evaluate.
        :param metadata (Optional[bool]): Optional boolean flag to return model name and version metadata in results.
        """
        if isinstance(targets, list):
            predict_targets = ", ".join([f'"{t}"' for t in targets])
        else:
            predict_targets = f'"{targets}"'

        # create list of properties
        properties = []
        if metadata:
            properties.append("METADATA")

        if len(properties) > 0:
            predict_targets += " WITH " + ", ".join(properties)

        # Add using query for current model and the latest version
        model_name = f'"{self.repo.name}" VERSION {self.version}'

        # Source can be a dataset, dataframe, or query string.
        source, params, connection_id = self._build_source(source, limit)

        # Set session connection from dataset if not set
        if not self.session.connection_id and self.dataset:
            self.session.set_connection(self.dataset.connection_id)
        query = f"EVALUATE {predict_targets} USING {model_name} GIVEN {source}"
        return self.session.execute(query, params=params, connection_id=connection_id)

    def predict(
        self,
        targets: Union[str, List[str], Dict[str, str]],
        source: Union[Dataset, pd.DataFrame, str],
        limit: Optional[int] = None,
        explanation: Optional[Union[bool, str]] = False,
        confidence: Optional[bool] = False,
        probabilities: Optional[Union[bool, List[any]]] = False,
        metadata: Optional[bool] = False,
        engine: Optional[Engine] = None,
    ) -> pd.DataFrame:
        """Make predictions for one or more target variables given a source dataset.

        :param targets: (Union[str, List[str], Dict[str, str]]): One or more target names to make predictions on.
        Include a dictionary with key being the target and the value being an alias to return in results.
        :param source: (Union[Dataset, pd.DataFrame, str]): Source to perform predictions on.
        :param limit: Number of rows to return.
        :param explanation: (Optional[Union[bool, str]]): Optional flag to enable explanations.
        Provide a string to return explanations for a specific algorithm.  Supported values are 'ig' or 'shap'.
        :param confidence: (Optional[bool]): Optional boolean flag to return metadata in results.
        :param probabilities: (Optional[Union[bool, List[any]]]): Optional flag to return probabilities for binary of
            multi-class classificaiton.
        Provide a list of category names eg [True,False] to include a specific set of classes.
        :param metadata: (Optional[bool]): Optional boolean flag to return model name and version metadata in results.
        :param engine: (Optional[Engine]): Option to select a particular engine for executing the predict.
        """

        # Add predict targets with optional alias, list, or as simple string
        if isinstance(targets, dict):
            predict_targets = ", ".join([f'"{k}" AS "{v}"' for (k, v) in targets.items()])
        elif isinstance(targets, list):
            predict_targets = ", ".join([f'"{t}"' for t in targets])
        else:
            predict_targets = f'"{targets}"'

        # create list of properties
        properties = []
        if confidence:
            properties.append("CONFIDENCE")
        if explanation:
            # TODO: Also add an enum type here for explanation instead of just string
            if isinstance(explanation, str):
                properties.append(f"EXPLANATION (algo='{explanation}')")
            else:
                properties.append("EXPLANATION")
        if probabilities:
            categories = ""
            if isinstance(probabilities, list):
                categories = ", ".join([f"'{str(c)}'" for c in probabilities])
                categories = f" OF ({categories})"
            properties.append("PROBABILITIES" + categories)
        if metadata:
            properties.append("METADATA")

        if len(properties) > 0:
            predict_targets += " WITH " + ", ".join(properties)

        # Add using query for current model and the latest version
        model_name = f'"{self.repo.name}" VERSION {self.version}'

        # Source can be a dataset, dataframe, or query string.
        source, params, connection_id = self._build_source(source, limit)
        # If using a dataframe, use model.dataset.connection_id
        if params:
            connection_id = self.dataset.connection_id

        query = f"PREDICT {predict_targets} USING {model_name} GIVEN {source}"
        return self.session.execute(
            query,
            params=params,
            connection_id=connection_id,
            engine_id=engine.id if engine else None,
        )

    def deploy(
        self,
        deployment_name: str,
        engine_template: Optional[str] = None,
        hf_token: Optional[str] = None,
        auto_suspend_seconds: Optional[int] = None,
        max_input_length: Optional[int] = None,
        max_total_tokens: Optional[int] = None,
        max_batch_prefill_tokens: Optional[int] = None,
        revision: Optional[str] = None,
        quantization_kwargs: Optional[Dict[str, str]] = None,
        overwrite: Optional[bool] = False,
    ) -> "interface.LLMDeploymentJob":
        if not self._llm_base_model_name:
            raise NotImplementedError(
                "Support for non-LLM deployments via this API is not yet implemented. "
                "Please use predibase_client.create_deployment() instead",
            )

        return interface.deploy_llm(
            session=self.session,
            deployment_name=deployment_name,
            engine_template=engine_template,
            hf_token=hf_token,
            auto_suspend_seconds=auto_suspend_seconds,
            max_input_length=max_input_length,
            max_total_tokens=max_total_tokens,
            max_batch_prefill_tokens=max_batch_prefill_tokens,
            quantization_kwargs=quantization_kwargs,
            revision=revision,
            predibase_model_uri=f"pb://models/{self.repo.name}/{self.version}",
            overwrite=overwrite,
        )

    def visualize(
        self,
        visualize_type: Union[VisualizeType, str],
        output_feature: Optional[str] = None,
        plot_images: Optional[bool] = True,
    ) -> List[str]:
        self._wait_until_ready()

        if isinstance(visualize_type, VisualizeType):
            visualize_type = visualize_type.value
        endpoint = f"/visualize?modelID={self.id}&visualizeType={visualize_type}"
        if output_feature:
            endpoint += f"&outputFeature={output_feature}"
        resp = self.session.get_json(endpoint)
        if plot_images:
            import ipyplot

            ipyplot.plot_images(
                resp["links"],
                img_width=400,
            )
        return resp["links"]

    def get_evaluation_statistics(self):
        self._wait_until_ready()
        endpoint = f"/visualize?modelID={self.id}&visualizeType=evaluation_statistics"
        resp = self.session.get_json(endpoint)
        return resp["testStatistics"]

    def get_feature_importance(self, output_feature: Optional[str] = None) -> pd.DataFrame:
        self._wait_until_ready()

        if output_feature is None:
            output_features = [x["name"] for x in self.config["output_features"]]
            if len(output_features) == 1:
                output_feature = output_features[0]
            else:
                raise ValueError(
                    "No output feature provided and "
                    "cannot be inferred because model contains multiple output features",
                )
        endpoint = f"/visualize/explanations?modelID={self.id}&outputFeature={output_feature}"
        resp = self.session.get_json(endpoint)
        return get_results_df({"dataset": resp})

    def get_training_timeline(self) -> pd.DataFrame:
        """
        :return: Dataframe with columns [Step, Start, End, Duration], where Step in
            {Preprocessing, Training, Evaluating, etc.}
        """
        endpoint = f"/models/version/{self.id}/timeline"
        resp = self.session.get_json_until(
            endpoint=endpoint,
            success_cond=lambda resp: all(
                [resp[process]["created"] is not None and resp[process]["completed"] is not None for process in resp],
            ),
        )
        columns = ["step", "error_message", "start", "end", "duration"]
        data = []
        for process in resp:
            start = parser.parse(resp[process]["created"])
            end = parser.parse(resp[process]["completed"])
            error_message = resp[process].get("errorMessage", None)
            data.append([process, error_message, start, end, end - start])
        return pd.DataFrame(columns=columns, data=data)

    def _build_source(
        self,
        source: Union[Dataset, pd.DataFrame, str],
        limit: Optional[int] = None,
    ) -> Tuple[str, Optional[str], Optional[str]]:
        # Source can be a dataset, dataframe, or query string.
        params = None
        connection_id = None
        if isinstance(source, Dataset):
            connection_id = source.connection_id
            source = f'SELECT * FROM "{source.connection.name}"."{source.name}"'
            if limit is not None:
                source += f" LIMIT {limit}"
        elif isinstance(source, pd.DataFrame):
            params = {"sdk_df_dataset": source.to_json()}
            source = "sdk_df_dataset"
        elif not isinstance(source, str):
            raise Exception(f"unexpected source type {type(source)}")
        return source, params, connection_id

    def to_draft(self):
        return ModelDraft(
            session=self.session,
            name=self.repo.name,
            repo_id=self.repo_id,
            _repo=self._repo,
            config=copy.deepcopy(self.config),
            dataset=self.dataset,
        )

    def fork(self, name):
        return ModelDraft(
            session=self.session,
            name=name,
            config=copy.deepcopy(self.config),
            dataset=self.dataset,
            parent_id=self.repo_id,
        )

    def get_active_fields(self) -> List[DataField]:
        """
        Queries like "PREDICT <field>> GIVEN SELECT * FROM <dataset>" that don't specify an explicit model
        will use this model for each of the fields returned here
        """
        return self.active_fields

    def activate_model_for_field(self, field: DataField):
        """Sets this model as the default model used for operations on a data field. For example, if you have two
        models.

        [model_tabnet, model_concat] trained on the same dataset, you can call
        model_tabnet.set_default_for_field(DataField(id=..., name="Survived")).

        Queries like "PREDICT <field>> GIVEN SELECT * FROM <dataset>" that don't specify an explicit model
        will use the default model set here for the "Survived" field

        :param field: DataField of dataset. Obtained from dataset.get_fields()
        """
        if field.id in [x.id for x in self.active_fields]:
            raise ValueError(f"field {field.name} is already active for this model")
        return self.session.post_json("/models/activate", {"modelID": self.id, "fieldIDs": [field.id]})

    def deactivate_model_for_field(self, field: DataField):
        if field.id not in [x.id for x in self.active_fields]:
            raise ValueError(f"field {field.name} is not active for this model")
        return self.session.post_json("/models/activate", {"modelID": self.id, "fieldIDs": [field.id]})

    def _wait_until_ready(self, launch_tensorboard=False):
        """Utility function to wait until a model is fully ready, including evaluation and visualization steps.
        Certain operations may otherwise fail in a racy way.

        :param model: the Model to wait for.
        :return: None
        """
        self.session._wait_until_model_ready_with_logging(self._id, launch_tensorboard)

    def launch_tensorboard(self):
        # Check if the model has logs available or will have logs available
        model_logs = self.session.get_model_logs_urls(self._id)
        if len(model_logs.get("logfiles", [])) == 0 and not model_logs.get("moreLogsAvailable", False):
            print(
                f"Model {self._id} does not have logs available, which is likely because the model failed or was terminated before logs were created. Please check model status.",  # noqa: E501
            )
            return

        # Sync files in subthread and launch TB in main thread
        tensorboard_event = Event()
        logs_sync_thread = Thread(
            target=self.session.get_model_logs,
            args=[self._id, tensorboard_event, None],
            daemon=True,
        )
        logs_sync_thread.start()
        launch_tb(self._id)

        # Stop syncing files when TB is stopped
        tensorboard_event.set()
        logs_sync_thread.join()


class ModelFuture:
    def __init__(self, version_id: int, session: Session):
        self.version_id = version_id
        self.session = session

    def wait(self):
        self._wait_for_ready()

    def get(
        self,
        launch_tensorboard: bool = False,
    ) -> Model:
        log_info("Monitoring status of model training...")

        for _ in range(5):
            resp = self.session.get_json(f"/models/version/{self.version_id}")
            engine_template = resp.get("modelVersion", {}).get("engineTemplate", None)

            if engine_template is None:
                time.sleep(1)
                continue

            cloud = next(iter(engine_template["headNodeGroup"]["nodeType"]["clouds"]))

            node_counts = defaultdict(int)
            head_node_group_cloud = engine_template["headNodeGroup"]["nodeType"]["clouds"][cloud]
            head_node_label = head_node_group_cloud.get("publicLabel", head_node_group_cloud["label"])
            node_counts[head_node_label] += 1
            for worker_node_group in engine_template.get("workerNodeGroups", []):
                public_label = (
                    worker_node_group.get("nodeType", {}).get("clouds", {}).get(cloud, {}).get("publicLabel", "")
                )
                label = (
                    public_label
                    if public_label
                    else worker_node_group.get("nodeType", {}).get("clouds", {}).get(cloud, {}).get("label", "")
                )
                count = int(
                    worker_node_group.get("replicas", "0"),
                )
                if label:
                    node_counts[label] += count

            log_info("Compute summary:")
            for label, count in sorted(node_counts.items()):
                if count > 0:
                    log_info(f"  * {label} (x{count})")

            break

        if launch_tensorboard:
            launch_tb(self.version_id)

        resp = self._wait_for_ready(
            launch_tensorboard=launch_tensorboard,
        )

        from predibase.resource_util import build_model

        return build_model(resp["modelVersion"], self.session)

    def cancel(self):
        self.session.post_json(f"/models/version/{self.version_id}/cancel", {})

    def is_finished(self) -> bool:
        resp = self.session.get_json(f"/models/version/{self.version_id}")
        return resp["modelVersion"]["status"] in {"ready", "failed", "canceled"}

    def _wait_for_ready(self, launch_tensorboard=False) -> Dict[str, Any]:
        return self.session._wait_until_model_ready_with_logging(self.version_id, launch_tensorboard)


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class ModelDraft:
    session: Session
    dataset: Dataset
    config: dict
    description: Optional[str] = None
    repo_id: Optional[int] = None
    _repo: Optional[ModelRepo] = None
    name: Optional[str] = None
    parent_id: Optional[str] = None

    def __repr__(self):
        repo_str = f"repo=({self.repo.name}...) " if self.repo else "repo=(<new repo from fork>)"
        return (
            f"ModelDraft(name={self.name}, dataset=({self.dataset.name}...), config={{...}}, "
            f"description={self.description}, {repo_str}, parent_id={self.parent_id})"
        )

    @property
    def repo(self):
        return self._repo

    # Removed https://github.com/predibase/predibase/pull/3770
    # @spinner(name="Train Model")
    def train(self, engine: Optional[Engine] = None, launch_tensorboard: bool = False) -> Model:
        if not self.session.is_plan_expired():
            future = self.train_async(engine)
            return future.get(launch_tensorboard=launch_tensorboard)
        else:
            raise PermissionError(
                "Training models is locked for expired plans. Contact us to upgrade.",
            )

    def train_async(self, engine: Optional[Engine] = None) -> ModelFuture:
        if not self.session.is_plan_expired():
            # Retraining
            if self.repo_id:
                request = {
                    "config": self.config,
                    "datasetID": self.dataset.id,
                    "repoID": self.repo_id,
                    "engineID": engine.id if engine is not None else None,
                    "description": self.description,
                }

                resp = self.session.post_json("/models/train", request)
                model_id = resp["model"]["id"]

                # Output link to UI model page
                endpoint = f"models/version/{model_id}"
                url = get_url(self.session, endpoint)
                log_info(f"Check Status of Model Training Here: [link={url}]{url}")

                return ModelFuture(model_id, self.session)

            # Forking
            endpoint = "/models/repo"
            request = {
                "modelType": "ludwig",
                "modelName": self.name,
                "latestConfig": self.config,
                "retrainCadence": "",
                "retrainConfig": {},
                "datasetID": self.dataset.id,
                "parentID": self.parent_id,
            }
            resp = self.session.post_json(endpoint, request)
            repo_id = resp["id"]
            log_info(f"Created model repository: <{self.name}>")

            resp = self.session.post_json(
                "/models/train",
                {
                    "datasetID": self.dataset.id,
                    "config": self.config,
                    "repoID": repo_id,
                    "engineID": engine.id if engine is not None else None,
                },
            )
            model_id, model_version = resp["model"]["id"], resp["model"]["repoVersion"]
            log_info(f"Training model version {model_version} for model repository <{self.name}>...")
            return ModelFuture(model_id, self.session)
        else:
            raise PermissionError(
                "Training models is locked for expired plans. Contact us to upgrade.",
            )


class FeatureMetrics:
    def __init__(self, values: Dict[str, float]):
        for k, v in values.items():
            setattr(self, k, v)


class SplitMetrics:
    def __init__(
        self,
        nested_metrics: Dict[str, Dict[str, float]],
        global_metrics: Dict[str, float],
    ):
        for k, v in nested_metrics.items():
            setattr(self, k, FeatureMetrics(v))
        for k, v in global_metrics.items():
            setattr(self, k, v)


class ModelMetrics:
    def __init__(
        self,
        nested_metrics: Dict[str, Dict[str, Dict[str, float]]],
        global_metrics: Dict[str, Dict[str, float]],
        general_metrics: Dict[str, int],
    ):
        # General metrics includes things like epochs, batch_size, etc.
        self.general_metrics = general_metrics
        self.train = SplitMetrics(nested_metrics[TRAIN_METRICS], global_metrics[TRAIN_METRICS])
        self.validation = SplitMetrics(nested_metrics[VALIDATION_METRICS], global_metrics[VALIDATION_METRICS])
        self.test = SplitMetrics(nested_metrics[TEST_METRICS], global_metrics[TEST_METRICS])


# TODO(travis): Consider moving these off of MLFlow format
@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class ModelRunMetric:
    key: str
    step: int
    timestamp: float
    value: float


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class ModelRunParam:
    key: str
    value: str


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class ModelRunTag:
    key: str
    value: str


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class ModelRunInfo:
    artifact_uri: str
    experiment_id: str
    lifecycle_stage: str  # active
    run_id: str
    run_uuid: str
    status: str
    user_id: str
    end_time: Optional[float] = field(metadata=config(field_name="EndTime"), default=None)
    start_time: Optional[float] = field(metadata=config(field_name="StartTime"), default=None)


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class ModelRunData:
    metrics: List[ModelRunMetric]
    params: List[ModelRunParam]
    tags: List[ModelRunTag]


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class ModelRun:
    # https://www.mlflow.org/docs/latest/_modules/mlflow/entities/run_data.html
    data: ModelRunData

    # https://www.mlflow.org/docs/latest/_modules/mlflow/entities/run_info.html#RunInfo
    info: ModelRunInfo


def get_suggested_configs(
    session: Session,
    base_config: Dict[Any, str],
) -> List[ConfigSuggestion]:
    """Gets config suggestions given a base config."""
    resp = session.post_json(
        "/config/suggest",
        {
            "config": base_config,
        },
    )
    config_suggestions = resp["configs"]
    return [ConfigSuggestion.from_dict(suggestion) for suggestion in config_suggestions]


def get_default_config(
    session: Session,
    dataset: Dataset,
    targets: Optional[Union[str, List[str]]] = None,
    automl: Optional[bool] = False,
    hyperopt: Optional[bool] = False,
) -> dict:
    endpoint = f"/config/detect/{dataset.id}"
    resp = session.get_json(endpoint)
    config = resp["config"]

    if targets:
        if isinstance(targets, str):
            targets = [targets]
        output_features = [x for x in config["input_features"] if x["name"] in targets]
        config["input_features"] = [x for x in config["input_features"] if x["name"] not in targets]
        config["output_features"] = output_features
    if not automl:
        if "combiner" in config:
            del config["combiner"]
        if "trainer" in config:
            del config["trainer"]
    if not hyperopt:
        if "hyperopt" in config:
            del config["hyperopt"]
    return config


def get_model(
    session: Session,
    name: Optional[str] = None,
    version: Optional[Union[int, str]] = None,
    model_id: Optional[int] = None,
) -> Model:
    if not (name or model_id):
        raise ValueError("Must provide either model_name or model_id")
    if name and model_id:
        raise ValueError("Cannot provide both model name and model_id")
    if version:
        if model_id:
            raise ValueError("Cannot provide both model version and model_id")

    if model_id:
        endpoint = f"/models/version/{model_id}"
        resp = session.get_json(endpoint)["modelVersion"]
    else:
        endpoint = f"/models/version/name/{name}"
        endpoint += f"?version={version}" if version else ""
        resp = session.get_json(endpoint)

    from predibase.resource_util import build_model

    return build_model(resp, session)


def get_model_repo(session: Session, name: str) -> ModelRepo:
    resp = session.get_json(f"/models/repo/name/{encode_url_param(name)}")

    return build_model_repo(resp, session)


def create_model_repo(
    session: Session,
    name: str,
    description: Optional[str] = None,
    engine: Optional[Engine] = None,
    exists_ok: bool = False,
) -> ModelRepo:
    if not session.is_plan_expired():
        try:
            session.post_json(
                "/models/repo",
                {
                    "modelName": name,
                    "description": description,
                    "modelType": "ludwig",
                    "retrainCadence": "",
                    "retrainConfig": {},
                    "parentID": None,
                    "selectedEngineID": engine.id if engine else None,
                },
            )
            log_info(f"Created model repository: <{name}>")
        except ServerResponseError as e:
            if exists_ok and e.code == 400:
                log_info(f"Model repository {name} already exists and new models will be added to it.")
            else:
                raise e

        # TODO(travis): would be more efficient to return model repo data in CREATE response.
        return get_model_repo(session, name)
    else:
        raise PermissionError(
            "Creating Model Repos is locked for expired plans. Contact us to upgrade or if you would like a demo",
        )



from __future__ import annotations

from typing import TYPE_CHECKING, Dict

from predibase._errors import PredibaseError
from predibase.resource.llm import interface
from predibase.resource.llm.interface import _LLMDeployment
from predibase.resources.model import FinetunedLLMAdapter, PretrainedHuggingFaceLLM

if TYPE_CHECKING:
    from predibase import Predibase

from predibase.resources.deployment import Deployment, ServerlessLLMDeployment, DedicatedLLMDeployment


class Deployments:
    def __init__(self, client: Predibase):
        self._client = client
        self._session = client._session

    def get(self, deployment_name: str) -> Deployment:
        # No need to provide the full old-style URI here since we know we're dealing with a Predibase deployment.
        if deployment_name.startswith("pb://deployments/"):
            deployment_name = deployment_name[len("pb://deployments/") :]
        if deployment_name.startswith("pb://"):
            deployment_name = deployment_name[len("pb://") :]

        # TODO: for now relying on the old payload and dataclass.
        data = _LLMDeployment.from_dict(self._session.get_json(f"/llms/{deployment_name}"))
        if data.is_shared:
            return ServerlessLLMDeployment(self._client, deployment_name)
        else:
            return DedicatedLLMDeployment(self._client, deployment_name)

    def create(
        self,
        name: str,
        model: FinetunedLLMAdapter | PretrainedHuggingFaceLLM,
        *,
        hf_token: str | None = None,
        auto_suspend_seconds: int | None = None,
        max_input_length: int | None = None,
        max_total_tokens: int | None = None,
        max_batch_prefill_tokens: int | None = None,
        quantization_kwargs: Dict[str, str] | None = None,
        revision: str | None = None,
        overwrite: bool | None = False,
        wait: bool = True,
    ) -> DedicatedLLMDeployment:
        if isinstance(model, FinetunedLLMAdapter):
            job = interface.deploy_llm(
                session=self._session,
                deployment_name=name,
                hf_token=hf_token,
                auto_suspend_seconds=auto_suspend_seconds,
                max_input_length=max_input_length,
                max_total_tokens=max_total_tokens,
                max_batch_prefill_tokens=max_batch_prefill_tokens,
                quantization_kwargs=quantization_kwargs,
                revision=revision,
                predibase_model_uri=model.uri,
                overwrite=overwrite,
            )
        elif isinstance(model, PretrainedHuggingFaceLLM):
            job = interface.deploy_llm(
                session=self._session,
                deployment_name=name,
                hf_token=hf_token,
                auto_suspend_seconds=auto_suspend_seconds,
                max_input_length=max_input_length,
                max_total_tokens=max_total_tokens,
                max_batch_prefill_tokens=max_batch_prefill_tokens,
                quantization_kwargs=quantization_kwargs,
                revision=revision,
                huggingface_model_name=model.name,
                overwrite=overwrite,
            )
        else:
            raise PredibaseError(f"unable to deploy unexpected model type {model.__class__}")

        deployment = DedicatedLLMDeployment(self._client, name)
        if wait:
            deployment.wait_ready()

        return deployment

    def delete(self, deployment: str | DedicatedLLMDeployment):
        if isinstance(deployment, str):
            deployment = DedicatedLLMDeployment(self._client, deployment)

        dep = deployment._refresh()
        if dep.data.is_shared():
            raise PredibaseError(f"cannot delete shared deployment {dep.name}")

        dep.delete()



from __future__ import annotations

from predibase.pql import Session
from predibase.resources.model import PretrainedHuggingFaceLLM


class Models:
    def __init__(self, session: Session):
        self._session = session

    def from_hf(self, model_name: str) -> PretrainedHuggingFaceLLM:
        return PretrainedHuggingFaceLLM(model_name)
